import os
import sys

from aim.web.configs import AIM_ENV_MODE_KEY


def build_db_upgrade_command():
    from aim import web
    web_dir = os.path.dirname(web.__file__)
    migrations_dir = os.path.join(web_dir, 'migrations')
    if os.getenv(AIM_ENV_MODE_KEY, 'prod') == 'prod':
        ini_file = os.path.join(migrations_dir, 'alembic.ini')
    else:
        ini_file = os.path.join(migrations_dir, 'alembic_dev.ini')
    return [sys.executable, '-m', 'alembic', '-c', ini_file, 'upgrade', 'head']


def build_uvicorn_command(host, port, num_workers, uds_path, ssl_keyfile, ssl_certfile, log_level):
    cmd = [sys.executable, '-m', 'uvicorn',
           '--host', host, '--port', f'{port}',
           '--workers', f'{num_workers}']
    if os.getenv(AIM_ENV_MODE_KEY, 'prod') == 'prod':
        log_level = log_level or 'error'
    else:
        import aim
        cmd += ['--reload', '--reload-dir', os.path.dirname(aim.__file__)]
        log_level = log_level or 'debug'
    if uds_path:
        cmd += ['--uds', uds_path]
    if ssl_keyfile:
        cmd += ['--ssl-keyfile', ssl_keyfile]
    if ssl_certfile:
        cmd += ['--ssl-certfile', ssl_certfile]
    cmd += ['--log-level', log_level.lower()]
    cmd += ['aim.web.run:app']
    return cmd


def get_free_port_num():
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    port_num = s.getsockname()[1]
    s.close()
    return port_num



import datetime

from multiprocessing import Process
from typing import Optional

import aim.ext.transport.remote_tracking_pb2 as rpc_messages
import aim.ext.transport.remote_tracking_pb2_grpc as remote_tracking_pb2_grpc
from aim.ext.cleanup import AutoClean


class RemoteWorkerAutoClean(AutoClean['RemoteWorker']):
    PRIORITY = 120

    def __init__(self, instance: 'RemoteWorker'):
        super().__init__(instance)
        self.worker_process = instance.worker_process

    def _close(self) -> None:
        self.worker_process.join()


class RemoteWorker:
    WORKER_START_GRACE_PERIOD = datetime.timedelta(minutes=5)

    def __init__(self, target_f, *, host, port, index, ssl_keyfile=None, ssl_certfile=None):
        self._resources: Optional[RemoteWorkerAutoClean] = None
        self.worker_process = Process(target=target_f, args=(host, port, ssl_keyfile, ssl_certfile))
        self._host = host
        self._port = port
        self._idx = index
        self._ssl_keyfile = ssl_keyfile
        self._ssl_certfile = ssl_certfile
        self._start_time = None
        self._clients = []
        self._resources = RemoteWorkerAutoClean(self)

    def start(self):
        self.worker_process.start()
        self._start_time = datetime.datetime.now()

    def stop(self):
        self.worker_process.join()

    def restart(self):
        if datetime.datetime.now() - self.start_time > self.WORKER_START_GRACE_PERIOD:
            self.stop()
            self.start()

    def add_client(self, client_uri):
        self._clients.append(client_uri)

    def remove_client(self, client_uri):
        if client_uri in self._clients:
            self._clients.remove(client_uri)

    @property
    def clients(self):
        return self._clients

    @property
    def start_time(self):
        return self._start_time

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return str(self._port)

    @property
    def address(self):
        return f'{self._host}:{self._port}'

    @property
    def client_count(self):
        return len(self._clients)

    @property
    def index(self):
        return str(self._idx)

    def cleanup_client_resources(self, client_uri):
        import grpc

        if self._ssl_certfile:
            with open(self._ssl_certfile, 'rb') as f:
                root_certificates = grpc.ssl_channel_credentials(f.read())
            _remote_worker_channel = grpc.secure_channel(self.address, root_certificates)
        else:
            _remote_worker_channel = grpc.insecure_channel(self.address)

        _remote_worker_stub = remote_tracking_pb2_grpc.RemoteTrackingServiceStub(_remote_worker_channel)
        request = rpc_messages.ClientResourceCleanupRequest(client_uri=client_uri)
        _remote_worker_stub.cleanup_client_resources(request)


class LocalWorker(RemoteWorker):
    def __init__(self, target_f, *, host, port, index, ssl_keyfile=None, ssl_certfile=None):
        self._resources: Optional[RemoteWorkerAutoClean] = None
        self._host = host
        self._port = port
        self._idx = index
        self._ssl_keyfile = ssl_keyfile
        self._ssl_certfile = ssl_certfile
        self._start_time = None
        self._clients = []

    def start(self):
        self._start_time = datetime.datetime.now()

    def stop(self):
        pass

    def restart(self):
        self.start()



import pytest

from tests.base import TestBase
from tests.utils import is_package_installed

from aim.sdk import Image
import numpy as np


class TestImageConstruction(TestBase):
        # greyscale, dims = (1,2,2)
    img1 = np.array([[[0, 1], [1, 0]]], np.uint8)
    # greyscale, dims = (2,2)
    img2 = np.array([[0, 1], [1, 0]], np.uint8)
    # RGB, dims = (3,2,2)
    img3 = np.array([[[0, 1], [1, 0]], [[0, 1], [1, 0]], [[0, 1], [1, 0]]], np.uint8)

    @pytest.mark.skipif(not is_package_installed('torch'), reason="'torch' is not installed. skipping.")
    def test_image_from_torch_tensor(self):
        import torch

        # order of channels in torch.tensor => channels, height, width

        self.assertEqual(Image(torch.tensor(self.img1)), Image(np.transpose(self.img1, (1, 2, 0))))
        self.assertEqual(Image(torch.tensor(self.img2)), Image(self.img2))
        self.assertEqual(Image(torch.tensor(self.img3)), Image(np.transpose(self.img3, (1, 2, 0))))

        self.assertEqual(Image(torch.tensor(self.img1.astype(np.float32))), Image(255 * np.transpose(self.img1, (1, 2, 0))))
        self.assertEqual(Image(torch.tensor(self.img2.astype(np.float32))), Image(255 * self.img2))
        self.assertEqual(Image(torch.tensor(self.img3.astype(np.float32))), Image(255 * np.transpose(self.img3, (1, 2, 0))))


    @pytest.mark.skipif(not is_package_installed('tensorflow'), reason="'tensorflow' is not installed. skipping.")
    def test_image_from_tf_tensor(self):
        import tensorflow as tf

        # order of channels in tf.tensor => height, width, channels

        self.assertEqual(Image(tf.convert_to_tensor(self.img1, tf.uint8)), Image(self.img1))
        self.assertEqual(Image(tf.convert_to_tensor(self.img2, tf.uint8)), Image(self.img2))
        self.assertEqual(Image(tf.convert_to_tensor(self.img3, tf.uint8)), Image(self.img3))

        self.assertEqual(Image(tf.convert_to_tensor(self.img1, tf.float32)), Image(255 * self.img1))
        self.assertEqual(Image(tf.convert_to_tensor(self.img2, tf.float32)), Image(255 * self.img2))
        self.assertEqual(Image(tf.convert_to_tensor(self.img3, tf.float32)), Image(255 * self.img3))


if __name__ == "__main__":
    TestImageConstruction().test_image_from_torch_tensor()
    TestImageConstruction().test_image_from_tf_tensor()



import os.path

import click
import shutil

from aim.cli.upgrade._legacy_repo import (
    AimRepo as LegacyRepo,
    Run as LegacyRun,
    deserialize_pb,
    AIM_MAP_METRICS_KEYWORD
)
from aim.sdk.configs import AIM_REPO_NAME
from aim.sdk.utils import clean_repo_path
from aim.sdk.run import Run
from aim.sdk.repo import Repo


class RepoIntegrityError(Exception):
    pass


def setup_directories(input_path: str):
    path = clean_repo_path(input_path)
    repo_path = path + '/.aim'
    new_repo_path = path + '/.aim_progress'
    l_repo_path = path + '/.aim_legacy'
    if not (os.path.exists(repo_path) and os.path.isdir(repo_path)):
        click.echo(f'\'{input_path}\' does not seem to be an Aim repository. Exiting.')
        exit(1)
    if os.path.exists(new_repo_path):
        shutil.rmtree(new_repo_path, ignore_errors=True)
    os.environ[AIM_REPO_NAME] = '.aim_progress'
    return repo_path, new_repo_path, l_repo_path


def collect_runs(lrepo: LegacyRepo):
    lruns = []
    for branch in lrepo.list_branches():
        for commit in lrepo.list_branch_commits(branch):
            lruns.append(LegacyRun(lrepo, experiment_name=branch, run_hash=commit))
    return lruns


def convert_run(lrun: LegacyRun, repo: Repo, legacy_run_map, skip_failed):
    try:
        run = Run(repo=repo,
                  system_tracking_interval=None)  # do not track system metrics as they already logged if needed

        lrun.open_storage()
        if lrun.params.get(AIM_MAP_METRICS_KEYWORD):
            del lrun.params[AIM_MAP_METRICS_KEYWORD]  # set internally. no need to copy
        run[...] = lrun.params
        run['v2_params'] = {'run_hash': lrun.run_hash}
        if 'process' in lrun.config:
            run['v2_params', 'start_date'] = lrun.config['process']['start_date']
            run['v2_params', 'finish_date'] = lrun.config['process']['finish_date']

        run.experiment = lrun.experiment_name
        if lrun.config.get('archived'):
            run.archived = True

        run_metrics = {}
        legacy_run_map[lrun.run_hash] = run_metrics
        for metric in lrun.get_all_metrics().values():
            try:
                metric.open_artifact()
                run_metrics[metric.name] = []
                for trace in metric.get_all_traces():
                    metric_name = metric.name
                    context = trace.context
                    run_metrics[metric.name].append(context)
                    for r in trace.read_records(slice(0, None, 1)):
                        step_record, metric_record = deserialize_pb(r)
                        # no need to track in a separate thread. use _track_impl directly.
                        run._track_impl(metric_record.value, step_record.timestamp, metric_name,
                                        step_record.step, step_record.epoch, context=context)
            except Exception:
                metric.close_artifact()
                raise
            finally:
                metric.close_artifact()
        del run
    except Exception as e:
        click.echo(f'\nFailed to convert run {lrun.run_hash}. Reason: {str(e)}.', err=True)
        if not skip_failed:
            raise
    finally:
        lrun.close_storage()


def check_repo_integrity(repo: Repo, legacy_run_map: dict):
    try:
        ok = True
        with click.progressbar(repo.iter_runs(),
                               item_show_func=lambda r: (f'Checking run {r.hash}' if r else '')) as runs:
            for run in runs:
                legacy_hash = run.get(['v2_params', 'run_hash'])
                run_metrics = legacy_run_map.pop(legacy_hash)
                for metric_name, ctx, _ in run.iter_metrics_info():
                    idx = run_metrics[metric_name].index(ctx.to_dict())
                    run_metrics[metric_name].pop(idx)
                    if not run_metrics[metric_name]:
                        del run_metrics[metric_name]
                if run_metrics:
                    click.echo(f'Run \'{run.hash}\' [\'{legacy_hash}\'] missing metrics \'{run_metrics.keys()}\'.')
                    ok = False
        if legacy_run_map:
            click.echo(f'Repo missing runs \'{legacy_run_map.keys()}\'.')
            ok = False
    except (KeyError, ValueError):
        ok = False
    if not ok:
        raise RepoIntegrityError


def convert_2to3(path: str, drop_existing: bool = False, skip_failed_runs: bool = False, skip_checks: bool = False):
    lrepo_path, repo_path, lrepo_dump_path = setup_directories(path)

    def _rollback():
        shutil.rmtree(repo_path, ignore_errors=True)
        if os.environ.get(AIM_REPO_NAME):
            del os.environ[AIM_REPO_NAME]

    try:
        click.echo('Preparing new repository...')
        lrepo = LegacyRepo(mode=LegacyRepo.READING_MODE, repo_full_path=lrepo_path)
        repo = Repo.from_path(repo_path, init=True)

        click.echo('Analyzing legacy repository...')
        lruns = collect_runs(lrepo)
        click.echo(f'Collected {len(lruns)} runs.')

        legacy_run_map = {}
        with repo.structured_db:
            with click.progressbar(
                    lruns, show_pos=True, item_show_func=lambda r: (f'Converting run {r.run_hash}' if r else '')
            ) as legacy_runs:
                for lrun in legacy_runs:
                    convert_run(lrun, repo, legacy_run_map, skip_failed=skip_failed_runs)

        if not skip_checks:
            click.echo('Checking repository integrity...')
            check_repo_integrity(repo, legacy_run_map)
            click.echo('Repository integrity check passed!')
        else:
            click.echo('WARN Skipping repository integrity checks.')
    except KeyboardInterrupt:
        click.echo('Convert interrupted by client. Rolling back...')
        _rollback()
        return
    except RepoIntegrityError:
        click.echo('Repository integrity check failed. Rolling back...', err=True)
        click.echo('If you want to convert repository anyway, please rerun command with \'--skip-checks\' flag.')
        _rollback()
        return
    except Exception as e:
        click.echo(f'Failed to convert repository. '
                   f'Reason: {str(e)}')
        _rollback()
        return
    if drop_existing:
        click.echo('Removing legacy repository...')
        shutil.rmtree(lrepo_path)
    else:
        shutil.move(lrepo_path, lrepo_dump_path)
        click.echo(f'Legacy repository can be found at \'{lrepo_dump_path}\'.')
    shutil.move(repo_path, lrepo_path)
    if os.environ.get(AIM_REPO_NAME):
        del os.environ[AIM_REPO_NAME]
    click.echo('DONE')
    return repo



from aim.storage.structured.entities import (
    ObjectFactory,
    Run, Experiment, Tag,
    RunCollection,
    ExperimentCollection,
    TagCollection,
)
from aim.storage.structured.sql_engine.entities import (
    ModelMappedRun,
    ModelMappedExperiment,
    ModelMappedTag,
)
from typing import List
from datetime import datetime


class ModelMappedFactory(ObjectFactory):
    def __init__(self):
        self._session_depth = 0
        self._session = None

    @staticmethod
    def run_cls():
        return ModelMappedRun

    @staticmethod
    def experiment_cls():
        return ModelMappedExperiment

    @staticmethod
    def tag_cls():
        return ModelMappedTag

    def runs(self) -> RunCollection:
        return ModelMappedRun.all(session=self._session or self.get_session())

    def search_runs(self, term: str) -> RunCollection:
        return ModelMappedRun.search(term, session=self._session or self.get_session())

    def find_run(self, _id: str) -> Run:
        return ModelMappedRun.find(_id, session=self._session or self.get_session())

    def find_runs(self, ids: List[str]) -> List[Run]:
        return ModelMappedRun.find_many(ids, session=self._session or self.get_session())

    def create_run(self, runhash: str, created_at: datetime = None) -> Run:
        run = ModelMappedRun.from_hash(runhash, created_at, session=self._session or self.get_session())
        run.experiment = 'default'
        return run

    def delete_run(self, runhash: str) -> bool:
        return ModelMappedRun.delete_run(runhash, session=self._session or self.get_session())

    def experiments(self) -> ExperimentCollection:
        return ModelMappedExperiment.all(session=self._session or self.get_session())

    def search_experiments(self, term: str) -> ExperimentCollection:
        return ModelMappedExperiment.search(term, session=self._session or self.get_session())

    def find_experiment(self, _id: str) -> Experiment:
        return ModelMappedExperiment.find(_id, session=self._session or self.get_session())

    def create_experiment(self, name: str) -> Experiment:
        return ModelMappedExperiment.from_name(name, session=self._session or self.get_session())

    def tags(self) -> TagCollection:
        return ModelMappedTag.all(session=self._session or self.get_session())

    def search_tags(self, term: str) -> TagCollection:
        return ModelMappedTag.search(term, session=self._session or self.get_session())

    def find_tag(self, _id: str) -> Tag:
        return ModelMappedTag.find(_id, session=self._session or self.get_session())

    def create_tag(self, name: str) -> Tag:
        return ModelMappedTag.from_name(name, session=self._session or self.get_session())

    def delete_tag(self, _id: str) -> bool:
        return ModelMappedTag.delete(_id, session=self._session or self.get_session())

    def get_session(self, autocommit=True):
        raise NotImplementedError

    def __enter__(self):
        if self._session_depth == 0:
            assert self._session is None
            self._session = self.get_session(autocommit=False)
        self._session_depth += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self._session_depth > 0
        self._session_depth -= 1
        if self._session_depth == 0:
            if exc_type is None:
                self._session.commit()
            else:
                self._session.rollback()
            self._session = None



import fnmatch
import os

from typing import List, Tuple
import io
import zipfile
from datetime import datetime

from aim.storage.rockscontainer import RocksContainer


def list_repo_runs(repo_path: str, lookup_dir: str = None) -> List[str]:
    if lookup_dir is None:
        chunks_dir = os.path.join(repo_path, '.aim', 'meta', 'chunks')
    else:
        chunks_dir = os.path.join(repo_path, '.aim', lookup_dir)
    return os.listdir(chunks_dir)


def match_runs(repo_path: str, hashes: List[str], lookup_dir: str = None) -> List[str]:
    matched_hashes = set()
    all_run_hashes = None
    for run_hash in hashes:
        if '*' in run_hash:
            expr = run_hash  # for the sake of readability
            # avoiding multiple or unnecessary list_runs() calls
            if not all_run_hashes:
                all_run_hashes = list_repo_runs(repo_path, lookup_dir)
            if expr == '*':
                return all_run_hashes
            # update the matches set with current expression matches
            matched_hashes.update(fnmatch.filter(all_run_hashes, expr))
        else:
            matched_hashes.add(run_hash)

    return list(matched_hashes)


def make_zip_archive(repo_path: str) -> io.BytesIO:
    aim_dir = os.path.join(repo_path, '.aim')
    zip_buf = io.BytesIO()
    zipf = zipfile.ZipFile(zip_buf, 'w', zipfile.ZIP_DEFLATED)
    len_dir_path = len(aim_dir)
    for root, _, files in os.walk(aim_dir):
        for file in files:
            file_path = os.path.join(root, file)
            zipf.write(file_path, file_path[len_dir_path:])
    zipf.close()
    return zip_buf


def upload_repo_runs(buffer: io.BytesIO, bucket_name: str) -> Tuple[bool, str]:
    try:
        import boto3
    except ImportError:
        raise RuntimeError(
            'This command requires \'boto3\' to be installed. '
            'Please install it with command: \n pip install boto3'
        )

    try:
        s3_client = boto3.client('s3')
        buckets = s3_client.list_buckets()
        bucket_names = []
        for bucket in buckets['Buckets']:
            bucket_names.append(bucket['Name'])

        if bucket_name not in bucket_names:
            s3_client.create_bucket(Bucket=bucket_name)

        key = f'aim-{datetime.now().strftime("%Y-%m-%d-%H-%M-%S")}.zip'
        s3_client.upload_fileobj(buffer, bucket_name, key)
        return True, key
    except Exception as e:
        return False, e


def optimize_container(path, extra_options):
    rc = RocksContainer(path, read_only=True, **extra_options)
    rc.optimize_for_read()



import os

from aim.cli.upgrade._legacy_repo.configs import AIM_OBJECTS_DIR_NAME, AIM_COMMIT_META_FILE_NAME


def get_experiment_path(repo_path: str, experiment_name: str) -> str:
    return os.path.join(repo_path, experiment_name)


def get_experiment_run_path(repo_path: str, experiment_name: str,
                            run_hash: str) -> str:
    path = os.path.join(get_experiment_path(repo_path, experiment_name),
                        run_hash)
    return path


def get_run_objects_dir_path(repo_path: str, experiment_name: str,
                             run_hash: str) -> str:
    """Returns path of `objects` directory of a run"""
    path = os.path.join(repo_path,
                        experiment_name,
                        run_hash,
                        AIM_OBJECTS_DIR_NAME)
    return path


def get_run_objects_meta_file_path(repo_path: str, experiment_name: str,
                                   run_hash: str) -> str:
    """Returns path of `meta.json` file of a run"""
    path = os.path.join(get_run_objects_dir_path(repo_path, experiment_name,
                                                 run_hash),
                        AIM_COMMIT_META_FILE_NAME)
    return path



import requests
from typing import Optional

from aim.ext.notifier.base_notifier import BaseNotifier


class WorkplaceNotifier(BaseNotifier):
    def __init__(self, _id: str, config: dict):
        super().__init__(_id)
        self.message_template = config['message']
        self.wp_access_token = config['access_token']
        self.wp_group_id = config['group_id']
        self.url = self._get_workplace_url()

    def notify(self, message: Optional[str] = None, **kwargs):
        message_template = message or self.message_template
        msg = message_template.format(**kwargs)
        params = {'access_token': self.wp_access_token}
        requests.post(self.url, json={'message': msg}, params=params)

    def _get_workplace_url(self):
        api_version = 'v14.0'

        return f'https://graph.facebook.com/{api_version}/{self.wp_group_id}/feed'



"""Hashing implementations of AimObject-compatible data.
This is different from CPython's implementation because of many reasons:
  *  CPython uses non-deterministic hashing to avoid collision attacks.
     Although we also cover such scenarios, in contrast to CPython, we do not
     use PYTHONHASHSEED environment variable.
  *  We implement hash for mutable objects as well (`list`, `dict`, etc.)
  *  Our implementation are less prone to manually designed collisions.
"""

import hashlib

from aim.storage.encoding import encode_int64, decode_int64  # noqa
from aim.storage.hashing import c_hash
from aim.storage.types import AimObject, AimObjectArray, AimObjectKey, AimObjectDict, NoneType

from typing import Tuple, Union


# We use 8-byte digest for Aim hashes
_HASH_SIZE = 8

# Here we fix hash seeds for primitive types
_HASH_NONE = 1393794177995070132

_HASH_UNIFORM_SALT = encode_int64(-7494597716789683807)
_HASH_ARRAY_SALT = encode_int64(2965039128708897743)
_HASH_OBJECT_SALT = encode_int64(-2435585119574290752)
_HASH_STR_SALT = encode_int64(7540324813251503183)
_HASH_BYTES_SALT = encode_int64(-6836296829636613855)


def hash_none(obj: NoneType = None) -> int:
    """Hash None values."""
    # We hard-code the hash value of None
    return _HASH_NONE


def hash_uniform(bad_hash: int) -> int:
    """Make a uniform permutation for mapping to make it more or less
    cryptographically stable, so it's more difficult to manually or (by chance
    in real applications) craft / find such examples that `a != b` but
    `hash(a) == hash(b)`
    """
    state = hashlib.blake2b(encode_int64(bad_hash),
                            digest_size=_HASH_SIZE,
                            salt=_HASH_UNIFORM_SALT)
    return decode_int64(state.digest())


def hash_number(number: Union[int, float]) -> int:
    """Hash a numeric value.

    As in Python, we make sure that `hash(a) == hash(b)` when `a == b` even if
    `a` and `b` are of different types (e.g. `int` vs `float`).
    In contrast to CPython implementation, our hash does not map integers to
    themselves making them more robust for different hasing scenarios.
    """
    state = c_hash.det_hash(number)
    return hash_uniform(state)


def hash_bool(obj: bool) -> int:
    """Hash a boolean value.

    We make sure that `hash(False) == hash(0) != hash(1) == hash(True)` similar
    to Python hash.
    """
    return hash_number(0) if not obj else hash_number(1)


def hash_bytes(obj: bytes) -> int:
    """Hash an `bytes` buffer"""
    # We use `blake2b` to hash the `bytes` object
    state = hashlib.blake2b(obj,
                            digest_size=_HASH_SIZE,
                            salt=_HASH_BYTES_SALT)
    return decode_int64(state.digest())


def hash_string(obj: str) -> int:
    """Hash an string object"""
    # Similar to `bytes`, we use `blake2b` to hash strings as well
    # First, we encode them to `utf-8` and then compute the hash
    # but *a different hash seed is provided* to make sure strings and their
    # utf-8 encoded blobs do not map to the same hash.
    state = hashlib.blake2b(obj.encode('utf-8'),
                            digest_size=_HASH_SIZE,
                            salt=_HASH_STR_SALT)
    return decode_int64(state.digest())


def hash_array(obj: AimObjectArray) -> int:
    """Hash an array of objects.

    We do not take into account whether it is a `list` or `tuple`, so
    `hash([1, 2, ['x', 5]]) == hash((1, 2, ('x', 5)))`
    """
    state = hashlib.blake2b(digest_size=_HASH_SIZE,
                            salt=_HASH_ARRAY_SALT)
    for i in obj:
        piece_hash = hash_auto(i)
        state.update(encode_int64(piece_hash))

    return decode_int64(state.digest())


def key_cmp(obj: Tuple[AimObjectKey, AimObject]):
    """Comparator for `(key, value)` pairs w.r.t. their hashes.
    This defines an (meaningless) order but ensures a deterministic one.
    """
    key, _ = obj
    return hash_auto(key)


def hash_object(obj: AimObjectDict) -> int:
    """Hash an dict-like object.

    The implementation does not take into account the order
    `hash({'a': 5, 'b': 7}) == hash({'b': 7, 'a': 5})`
    """
    state = hashlib.blake2b(digest_size=_HASH_SIZE,
                            salt=_HASH_OBJECT_SALT)
    # Here we use `key_cmp` to run over the object keys in an (meaningless but)
    # deterministic order.
    for key_val_tuple in sorted(obj.items(), key=key_cmp):
        # We accumulate the hash state with `(key, value)`, more specificially
        # the array-hash of `(key, value)`
        piece_hash = hash_array(key_val_tuple)
        state.update(encode_int64(piece_hash))

    return decode_int64(state.digest())


def hash_auto(obj: AimObject) -> int:
    """Automatically detect and hash any AimObject-compatible object.

    The hash is an signed 64-bit integer.
    """
    if obj is None:
        return hash_none()
    elif isinstance(obj, bool):
        return hash_bool(obj)
    elif isinstance(obj, int):
        # We use same hash function for all the numbers
        return hash_number(obj)
    elif isinstance(obj, float):
        # We use same hash function for all the numbers
        return hash_number(obj)
    elif isinstance(obj, bytes):
        return hash_bytes(obj)
    elif isinstance(obj, str):
        return hash_string(obj)
    elif isinstance(obj, (tuple, list)):
        # Both `tuple` and `list` types are hashed through `hash_array`
        return hash_array(obj)
    elif isinstance(obj, dict):
        return hash_object(obj)
    else:
        raise NotImplementedError



from typing import Optional

from aim.sdk.legacy.session.session import get_default_session
from aim.sdk.legacy.deprecation_warning import deprecated


@deprecated
def track(*args, **kwargs):
    sess = get_default_session()
    return sess.track(*args, **kwargs)


@deprecated
def set_params(params: dict, name: Optional[str] = None):
    sess = get_default_session()
    return sess.set_params(params, name)



import time
import datetime
import pytz
import logging
import os

from threading import Thread
from pathlib import Path

from typing import Iterable


from aim.sdk.repo import Repo
from aim.sdk.run_status_watcher import Event, GRACE_PERIOD

logger = logging.getLogger(__name__)


class RepoIndexManager:
    index_manager_pool = {}

    @classmethod
    def get_index_manager(cls, repo: Repo):
        mng = cls.index_manager_pool.get(repo.path, None)
        if mng is None:
            mng = RepoIndexManager(repo)
            cls.index_manager_pool[repo.path] = mng
        return mng

    def __init__(self, repo: Repo):
        self.repo_path = repo.path
        self.repo = repo
        self.progress_dir = Path(self.repo_path) / 'meta' / 'progress'
        self.progress_dir.mkdir(parents=True, exist_ok=True)

        self.heartbeat_dir = Path(self.repo_path) / 'check_ins'
        self.run_heartbeat_cache = {}

        self._indexing_in_progress = False
        self._reindex_thread: Thread = None

    @property
    def repo_status(self):
        if self._indexing_in_progress is True:
            return 'indexing in progress'
        if self.reindex_needed:
            return 'needs indexing'
        return 'up-to-date'

    @property
    def reindex_needed(self) -> bool:
        runs_with_progress = os.listdir(self.progress_dir)
        return len(runs_with_progress) > 0

    def start_indexing_thread(self):
        logger.info(f'Starting indexing thread for repo \'{self.repo_path}\'')
        self._reindex_thread = Thread(target=self._run_forever, daemon=True)
        self._reindex_thread.start()

    def _run_forever(self):
        idle_cycles = 0
        while True:
            self._indexing_in_progress = False
            for run_hash in self._next_stalled_run():
                logger.info(f'Found un-indexed run {run_hash}. Indexing...')
                self._indexing_in_progress = True
                idle_cycles = 0
                try:
                    self.index(run_hash)
                except Exception as e:
                    logger.warning(f'Failed to index Run \'{run_hash}\'. Error: {e}.')
                    pass

                # sleep for small interval to release index db lock in between and allow
                # other running jobs to properly finalize and index Run.
                sleep_interval = .1
                time.sleep(sleep_interval)
            if not self._indexing_in_progress:
                idle_cycles += 1
                sleep_interval = 2 * idle_cycles if idle_cycles < 5 else 10
                logger.info(f'No un-indexed runs found. Next check will run in {sleep_interval} seconds. '
                            f'Waiting for un-indexed run...')
                time.sleep(sleep_interval)

    def _runs_with_progress(self) -> Iterable[str]:
        runs_with_progress = os.listdir(self.progress_dir)
        run_hashes = sorted(runs_with_progress, key=lambda r: os.path.getmtime(os.path.join(self.progress_dir, r)))
        return run_hashes

    def _next_stalled_run(self):
        for run_hash in self._runs_with_progress():
            if self._is_run_stalled(run_hash):
                yield run_hash

    def _is_run_stalled(self, run_hash: str) -> bool:
        heartbeat_files = list(sorted(self.heartbeat_dir.glob(f'{run_hash}-*-progress-*-*'), reverse=True))
        if heartbeat_files:
            last_heartbeat = Event(heartbeat_files[0].name)
            last_recorded_heartbeat = self.run_heartbeat_cache.get(run_hash)
            if last_recorded_heartbeat is None:
                self.run_heartbeat_cache[run_hash] = last_heartbeat
            else:
                if last_heartbeat.idx > last_recorded_heartbeat.idx:
                    self.run_heartbeat_cache[run_hash] = last_heartbeat
                else:
                    time_passed = time.time() - last_recorded_heartbeat.detected_epoch_time
                    if last_recorded_heartbeat.next_event_in + GRACE_PERIOD < time_passed:
                        return True

    def run_needs_indexing(self, run_hash: str) -> bool:
        return os.path.exists(self.progress_dir / run_hash)

    def index(self, run_hash) -> bool:
        try:
            index = self.repo._get_index_tree('meta', timeout=10).view(())
            meta_tree = self.repo.request_tree('meta', run_hash, read_only=False).subtree('meta')
            meta_run_tree = meta_tree.subtree('chunks').subtree(run_hash)
            meta_run_tree.finalize(index=index)
            if meta_run_tree['end_time'] is None:
                index['meta', 'chunks', run_hash, 'end_time'] = datetime.datetime.now(pytz.utc).timestamp()
            return True
        except TimeoutError:
            logger.warning(f'Cannot index Run {run_hash}. Index is locked.')
            return False



from pydantic import BaseModel
from typing import Dict, List, Optional


class ProjectApiOut(BaseModel):
    name: str
    path: str
    description: str
    telemetry_enabled: int


class ProjectParamsOut(BaseModel):
    params: Optional[Dict] = None
    metric: Optional[Dict[str, list]] = None
    images: Optional[Dict[str, list]] = None
    texts: Optional[Dict[str, list]] = None
    figures: Optional[Dict[str, list]] = None
    distributions: Optional[Dict[str, list]] = None
    audios: Optional[Dict[str, list]] = None


class ProjectActivityApiOut(BaseModel):
    num_experiments: int
    num_runs: int
    num_archived_runs: int
    num_active_runs: int
    activity_map: Dict[str, int] = {"2021-01-01": 54}


class Sequence(BaseModel):
    name: str
    context: dict


class ProjectPinnedSequencesApiOut(BaseModel):
    sequences: List[Sequence] = []


class ProjectPinnedSequencesApiIn(BaseModel):
    sequences: List[Sequence]



import datetime
import logging
import time
import weakref

from collections import Counter
from threading import Thread
from typing import Union

import aim.ext.transport.remote_router_pb2 as rpc_messages

logger = logging.getLogger(__name__)


class RPCHeartbeatSender(object):

    HEARTBEAT_INTERVAL_DEFAULT = 10
    NETWORK_CHECK_INTERVAL = 180

    NETWORK_UNSTABLE_WARNING_TEMPLATE = 'Network connection between client `{}` ' \
                                        'and server `{}` appears to be unstable.'
    NETWORK_ABSENT_WARNING_TEMPLATE = 'Network connection between client `{}` ' \
                                      'and server `{}` appears to be absent.'

    def __init__(self,
                 client,
                 interval: Union[int, float] = HEARTBEAT_INTERVAL_DEFAULT,
                 ):
        self._remote_client = weakref.ref(client)
        self._heartbeat_send_interval = interval

        # network state check vars
        self._network_stability_check_interval = RPCHeartbeatSender.NETWORK_CHECK_INTERVAL
        self._network_unstable_warned = False
        self._network_absent_warned = False
        self._heartbeat_responses = Counter(success=0, fail=0)

        # Start thread to collect stats and logs at intervals
        self._th_collector = Thread(target=self._target_f, daemon=True)
        self._shutdown = False
        self._started = False

    def start(self):
        if self._started:
            return

        self._started = True
        self._th_collector.start()

    def stop(self):
        if not self._started:
            return

        self._shutdown = True
        self._th_collector.join()

    def _send_heartbeat(self):
        if self._remote_client():
            try:
                response = self._remote_client().client_heartbeat()
                if response.status == rpc_messages.HeartbeatResponse.Status.OK:
                    self._heartbeat_responses['success'] += 1
                else:
                    self._heartbeat_responses['fail'] += 1
            except Exception:
                # at the moment we don't care about failures for heartbeats
                self._heartbeat_responses['fail'] += 1

    def _target_f(self):
        heartbeat_interval_counter = 0
        stability_check_interval_counter = 0
        # send initial heartbeat
        self._send_heartbeat()

        while True:
            if self._shutdown:
                break

            time.sleep(1)
            heartbeat_interval_counter += 1
            stability_check_interval_counter += 1

            if heartbeat_interval_counter > self._heartbeat_send_interval:
                self._send_heartbeat()
                heartbeat_interval_counter = 0

            if stability_check_interval_counter > self._network_stability_check_interval:
                self._check_network_state()
                stability_check_interval_counter = 0

    def _check_network_state(self):
        def reset_responses():
            self._heartbeat_responses['fail'] = 0
            self._heartbeat_responses['success'] = 0

        if not self._heartbeat_responses['fail']:
            reset_responses()
            return

        if self._heartbeat_responses['success'] and not self._network_unstable_warned:
            self._network_unstable_warned = True
            logger.warning(RPCHeartbeatSender.NETWORK_UNSTABLE_WARNING_TEMPLATE
                           .format(self._remote_client().uri, self._remote_client().remote_path))
            reset_responses()
            return

        if not self._network_absent_warned:
            self._network_absent_warned = True
            logger.warning(RPCHeartbeatSender.NETWORK_ABSENT_WARNING_TEMPLATE
                           .format(self._remote_client().uri, self._remote_client().remote_path))

        reset_responses()


class RPCHeartbeatWatcher:
    CLIENT_KEEP_ALIVE_TIME_DEFAULT = 30 * 60  # 30 minutes

    def __init__(self,
                 heartbeat_pool,
                 worker_pool,
                 keep_alive_time: Union[int, float] = CLIENT_KEEP_ALIVE_TIME_DEFAULT):

        self._heartbeat_pool = heartbeat_pool
        self._worker_pool = worker_pool

        self._client_keep_alive_time = keep_alive_time

        # Start thread to collect stats and logs at intervals
        self._th_collector = Thread(target=self._interval_check, daemon=True)
        self._shutdown = False
        self._started = False

    def start(self):
        if self._started:
            return

        self._started = True
        self._th_collector.start()

    def stop(self):
        if not self._started:
            return

        self._shutdown = True
        self._th_collector.join()

    def _release_client_resources(self, dead_client_uri):
        logger.warning(f'Cleaning up resources for client `{dead_client_uri}`.')
        for worker in self._worker_pool:
            if dead_client_uri in worker.clients:
                worker.cleanup_client_resources(dead_client_uri)
                worker.remove_client(dead_client_uri)

    def _interval_check(self):
        while True:
            # Get system statistics
            if self._shutdown:
                break

            time.sleep(1)

            client_uris = list(self._heartbeat_pool.keys())
            for client_uri in client_uris:
                now = datetime.datetime.now().timestamp()
                if now - self._heartbeat_pool[client_uri] > self._client_keep_alive_time:
                    self._release_client_resources(client_uri)
                    del self._heartbeat_pool[client_uri]



from pydantic import BaseModel
from datetime import datetime
from typing import List, Optional
from uuid import UUID


class ExploreStateCreateIn(BaseModel):
    type: str
    state: dict


class ExploreStateUpdateIn(BaseModel):
    type: Optional[str] = None
    state: Optional[dict] = None


class ExploreStateGetOut(BaseModel):
    id: UUID
    type: str
    updated_at: datetime = 'Wed, 01 Jan 2021 16:12:07 GMT'
    created_at: datetime = 'Wed, 01 Jan 2021 16:12:07 GMT'
    state: dict


ExploreStateListOut = List[ExploreStateGetOut]



from typing import Union, Tuple

from aim.sdk.sequence import MediaSequenceBase
from aim.sdk.objects.image import Image


class Images(MediaSequenceBase):
    """Class representing series of Image objects or Image lists."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        image_typename = Image.get_typename()
        return image_typename, f'list({image_typename})'

    @classmethod
    def sequence_name(cls) -> str:
        return 'images'



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: metric.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='metric.proto',
  package='artifacts.proto',
  syntax='proto3',
  serialized_pb=_b('\n\x0cmetric.proto\x12\x0f\x61rtifacts.proto\"\x1d\n\x0cMetricRecord\x12\r\n\x05value\x18\x01 \x01(\x01\x62\x06proto3')
)




_METRICRECORD = _descriptor.Descriptor(
  name='MetricRecord',
  full_name='artifacts.proto.MetricRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='artifacts.proto.MetricRecord.value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=33,
  serialized_end=62,
)

DESCRIPTOR.message_types_by_name['MetricRecord'] = _METRICRECORD
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MetricRecord = _reflection.GeneratedProtocolMessageType('MetricRecord', (_message.Message,), dict(
  DESCRIPTOR = _METRICRECORD,
  __module__ = 'metric_pb2'
  # @@protoc_insertion_point(class_scope:artifacts.proto.MetricRecord)
  ))
_sym_db.RegisterMessage(MetricRecord)


# @@protoc_insertion_point(module_scope)



import time

from tests.base import TestBase

from aim.sdk.types import QueryReportMode
from aim.sdk import Run


class TestRunResourceTracker(TestBase):
    def test_default_tracking_interval(self):
        run = Run()  # default tracking interval 10s
        run_hash = run.hash
        run.track(1, name='metric')
        time.sleep(.1)  # allow tracker to add resource usage metrics
        del run

        metrics = list(self.repo.query_metrics(f'run.hash == "{run_hash}" and metric.name.startswith("__")',
                                               report_mode=QueryReportMode.DISABLED))
        expected_metrics = {'__system__cpu', '__system__disk_percent',
                            '__system__memory_percent', '__system__p_memory_percent'}
        metric_names = set(m.name for m in metrics)
        for name in expected_metrics:
            self.assertIn(name, metric_names)

    def test_custom_tracking_interval(self):
        run = Run(system_tracking_interval=1)
        run_hash = run.hash
        run.track(1, name='metric')
        time.sleep(3)  # allow tracker to add resource usage metrics
        del run

        metrics = list(self.repo.query_metrics(f'run.hash == "{run_hash}" and metric.name.startswith("__")',
                                               report_mode=QueryReportMode.DISABLED))
        expected_metrics = {'__system__cpu', '__system__disk_percent',
                            '__system__memory_percent', '__system__p_memory_percent'}
        metric_names = set(m.name for m in metrics)
        for name in expected_metrics:
            self.assertIn(name, metric_names)
        for metric in metrics:
            # 3 sec. runtime, 1 sec. interval
            self.assertGreaterEqual(len(metric.data.indices_list()), 3)
            self.assertLessEqual(len(metric.data.indices_list()), 4)

    def test_disable_resource_tracking(self):
        run = Run(system_tracking_interval=None)
        run_hash = run.hash
        run.track(1, name='metric')
        time.sleep(.1)  # allow tracker to add resource usage metrics
        del run

        metrics = list(self.repo.query_metrics(f'run.hash == "{run_hash}" and metric.name.startswith("__")',
                                               report_mode=QueryReportMode.DISABLED))
        self.assertListEqual([], metrics)

    def test_resource_tracking_interval_limits(self):
        run = Run(system_tracking_interval=0, capture_terminal_logs=False)
        self.assertIsNone(run._system_resource_tracker)
        run = Run(system_tracking_interval=2 * 24 * 3600, capture_terminal_logs=False)  # two days
        self.assertIsNone(run._system_resource_tracker)



# The module is heavily based on `wrapt` package sources
# See: https://github.com/GrahamDumpleton/wrapt/blob/develop/src/wrapt/wrappers.py

from functools import lru_cache
import operator


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    return meta("NewBase", bases, {})


class _ObjectProxyMethods(object):

    # We use properties to override the values of __module__ and
    # __doc__. If we add these in ObjectProxy, the derived class
    # __dict__ will still be setup to have string variants of these
    # attributes and the rules of descriptors means that they appear to
    # take precedence over the properties in the base class. To avoid
    # that, we copy the properties into the derived class type itself
    # via a meta class. In that way the properties will always take
    # precedence.

    @property
    def __module__(self):
        return self.__wrapped__().__module__

    @__module__.setter
    def __module__(self, value):
        self.__wrapped__().__module__ = value

    @property
    def __doc__(self):
        return self.__wrapped__().__doc__

    @__doc__.setter
    def __doc__(self, value):
        self.__wrapped__().__doc__ = value

    # We similar use a property for __dict__. We need __dict__ to be
    # explicit to ensure that vars() works as expected.

    @property
    def __dict__(self):
        return self.__wrapped__().__dict__

    # Need to also propagate the special __weakref__ attribute for case
    # where decorating classes which will define this. If do not define
    # it and use a function like inspect.getmembers() on a decorator
    # class it will fail. This can't be in the derived classes.

    @property
    def __weakref__(self):
        return self.__wrapped__().__weakref__


class _ObjectProxyMetaType(type):
    def __new__(cls, name, bases, dictionary):
        # Copy our special properties into the class so that they
        # always take precedence over attributes of the same name added
        # during construction of a derived class. This is to save
        # duplicating the implementation for them in all derived classes.

        dictionary.update(vars(_ObjectProxyMethods))

        return type.__new__(cls, name, bases, dictionary)


class Eager1:

    __slots__ = ('name', 'wrapped')

    def __init__(self, wrapped, name):
        self.wrapped = lru_cache()(wrapped)
        self.name = name

    def __call__(self):
        w = self.wrapped()
        attr = getattr(w, self.name, Undefined())

        if not isinstance(attr, Undefined):
            return attr

        try:
            item = w[self.name]
            return item
        except KeyError:
            pass

        return Undefined()


class Eager2:

    __slots__ = ('view', 'name', 'cache')

    def __init__(self, view, name, cache):
        self.view = view
        self.name = name
        self.cache = cache

    def __call__(self):
        res = None
        cache_key = None
        if self.cache is not None:
            try:
                cache_key = self.view.container.prefix + str(self.name).encode()
                res = self.cache.get(cache_key)
            except Exception:
                pass

        if res is None:
            w = self.view
            res = getattr(w, self.name, Undefined())

            if not isinstance(res, Undefined):
                return res

            try:
                item = w[self.name]
                res = item
            except KeyError:
                res = Undefined()

            if self.cache is not None and cache_key is not None:
                self.cache[cache_key] = res

        return res


class Eager3:

    __slots__ = ('wrapped', 'key')

    def __init__(self, wrapped, key):
        self.wrapped = lru_cache()(wrapped)
        self.key = key

    def __call__(self):
        w = self.wrapped()

        try:
            item = w[self.key]
            return item
        except KeyError:
            pass

        return Undefined()


class Eager4:

    __slots__ = ('view', 'key', 'cache')

    def __init__(self, view, key, cache):
        self.view = view
        self.key = key
        self.cache = cache

    def __call__(self):
        res = None
        cache_key = None
        if self.cache is not None:
            try:
                cache_key = self.view.container.prefix + str(self.key).encode()
                res = self.cache.get(cache_key)
            except Exception:
                pass

        if res is None:
            w = self.view

            try:
                res = w[self.key]
            except KeyError:
                res = Undefined()

            if self.cache is not None and cache_key is not None:
                self.cache[cache_key] = res

            return res


class AimObjectProxy(with_metaclass(_ObjectProxyMetaType)):

    __slots__ = ('__wrapped__', '__view__')

    @property
    def __name__(self):
        return self.__wrapped__().__name__

    @__name__.setter
    def __name__(self, value):
        self.__wrapped__().__name__ = value

    @property
    def __class__(self):
        return self.__wrapped__().__class__

    @__class__.setter # noqa
    def __class__(self, value):  # noqa
        self.__wrapped__().__class__ = value

    @property
    def __annotations__(self):
        return self.__wrapped__().__annotations__

    @__annotations__.setter
    def __annotations__(self, value):
        self.__wrapped__().__annotations__ = value

    def __dir__(self):
        return dir(self.__wrapped__())

    def __str__(self):
        return str(self.__wrapped__())

    def __bytes__(self):
        return bytes(self.__wrapped__())

    def __repr__(self):
        return '<{} at 0x{:x} for {} at 0x{:x}>'.format(
            type(self).__name__, id(self),
            type(self.__wrapped__()).__name__,
            id(self.__wrapped__()))

    def __reversed__(self):
        return reversed(self.__wrapped__())

    def __round__(self):
        return round(self.__wrapped__())

    def __mro_entries__(self, bases):
        return (self.__wrapped__(),)

    def __lt__(self, other):
        try:
            return self.__wrapped__() < other
        except TypeError:
            return False

    def __le__(self, other):
        try:
            return self.__wrapped__() <= other
        except TypeError:
            return False

    def __eq__(self, other):
        return self.__wrapped__() == other

    def __ne__(self, other):
        return self.__wrapped__() != other

    def __gt__(self, other):
        try:
            return self.__wrapped__() > other
        except TypeError:
            return False

    def __ge__(self, other):
        try:
            return self.__wrapped__() >= other
        except TypeError:
            return False

    def __hash__(self):
        return hash(self.__wrapped__())

    def __nonzero__(self):
        return bool(self.__wrapped__())

    def __bool__(self):
        return bool(self.__wrapped__())

    def __add__(self, other):
        return self.__wrapped__() + other

    def __sub__(self, other):
        return self.__wrapped__() - other

    def __mul__(self, other):
        return self.__wrapped__() * other

    def __div__(self, other):
        return operator.div(self.__wrapped__(), other)

    def __truediv__(self, other):
        return operator.truediv(self.__wrapped__(), other)

    def __floordiv__(self, other):
        return self.__wrapped__() // other

    def __mod__(self, other):
        return self.__wrapped__() % other

    def __divmod__(self, other):
        return divmod(self.__wrapped__(), other)

    def __pow__(self, other, *args):
        return pow(self.__wrapped__(), other, *args)

    def __lshift__(self, other):
        return self.__wrapped__() << other

    def __rshift__(self, other):
        return self.__wrapped__() >> other

    def __and__(self, other):
        return self.__wrapped__() & other

    def __xor__(self, other):
        return self.__wrapped__() ^ other

    def __or__(self, other):
        return self.__wrapped__() | other

    def __radd__(self, other):
        return other + self.__wrapped__()

    def __rsub__(self, other):
        return other - self.__wrapped__()

    def __rmul__(self, other):
        return other * self.__wrapped__()

    def __rdiv__(self, other):
        return operator.div(other, self.__wrapped__())

    def __rtruediv__(self, other):
        return operator.truediv(other, self.__wrapped__())

    def __rfloordiv__(self, other):
        return other // self.__wrapped__()

    def __rmod__(self, other):
        return other % self.__wrapped__()

    def __rdivmod__(self, other):
        return divmod(other, self.__wrapped__())

    def __rpow__(self, other, *args):
        return pow(other, self.__wrapped__(), *args)

    def __rlshift__(self, other):
        return other << self.__wrapped__()

    def __rrshift__(self, other):
        return other >> self.__wrapped__()

    def __rand__(self, other):
        return other & self.__wrapped__()

    def __rxor__(self, other):
        return other ^ self.__wrapped__()

    def __ror__(self, other):
        return other | self.__wrapped__()

    def __neg__(self):
        return -self.__wrapped__()

    def __pos__(self):
        return +self.__wrapped__()

    def __abs__(self):
        return abs(self.__wrapped__())

    def __invert__(self):
        return ~self.__wrapped__()

    def __int__(self):
        return int(self.__wrapped__())

    def __float__(self):
        return float(self.__wrapped__())

    def __complex__(self):
        return complex(self.__wrapped__())

    def __oct__(self):
        return oct(self.__wrapped__())

    def __hex__(self):
        return hex(self.__wrapped__())

    def __index__(self):
        return operator.index(self.__wrapped__())

    def __len__(self):
        return len(self.__wrapped__())

    def __contains__(self, value):
        return value in self.__wrapped__()

    def __iter__(self):
        return iter(self.__wrapped__())

    __slots__ = '__wrapped__', '__view__', 'cache'

    def __init__(self, wrapped, view=None, cache=None):
        object.__setattr__(self, '__wrapped__', wrapped)
        # print('__wrapped__()lazy__', wrapped)
        object.__setattr__(self, '__view__', view)

        # Python 3.2+ has the __qualname__ attribute, but it does not
        # allow it to be overridden using a property and it must instead
        # be an actual string object instead.
        self.cache = cache
        try:
            object.__setattr__(self, '__qualname__', wrapped.__qualname__)
        except AttributeError:
            pass

    def __getattr__(self, name):
        # If we are being to lookup '__wrapped__()' then the
        # '__init__()' method cannot have been called.

        if name == '__wrapped__':
            raise ValueError('wrapper has not been initialised')

        if self.__view__ is None:
            return AimObjectProxy(Eager1(self.__wrapped__, name))

        return AimObjectProxy(
            Eager2(self.__view__, name, self.cache),
            self.__view__.view(name),
            self.cache
        )

    def __getitem__(self, key):

        if self.__view__ is None:
            return AimObjectProxy(Eager3(self.__wrapped__, key))

        return AimObjectProxy(
            Eager4(self.__view__, key, self.cache),
            self.__view__.view(key),
            self.cache
        )

    def __call__(self, *args, **kwargs):
        return self.__wrapped__()(*args, **kwargs)


class Undefined:
    def __bool__(self) -> bool:
        return False



import sys
import os
import platform

from shutil import rmtree
from setuptools import find_packages, setup, Command, Extension
from Cython.Build import cythonize
from aimrocks import lib_utils
# TODO This `setup.py` assumes that `Cython` and `aimrocks` are installed.
# This is okay for now as users are expected to install `aim` from wheels.

version_file = 'aim/VERSION'

with open(version_file) as vf:
    __version__ = vf.read().strip()

here = os.path.abspath(os.path.dirname(__file__))

# Package meta-data.
NAME = 'aim-with-auth-support'
DESCRIPTION = 'A super-easy way to record, search and compare AI experiments.'
VERSION = __version__
REQUIRES_PYTHON = '>=3.6.0'

# Get packages
packages = find_packages(exclude=('tests', 'aim.web.ui'))


# Get a list of all files in the html directory to include in our module
def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


migration_files = package_files('aim/web/migrations')
storage_migration_files = package_files('aim/storage/migrations')
notifier_files = package_files('aim/ext/notifier')
version_files = ['../aim/VERSION', ]

readme_file = 'README.md'
readme_text = open('/'.join((here, readme_file)), encoding="utf-8").read()
LONG_DESCRIPTION = readme_text.strip()

SETUP_REQUIRED = [
    'Cython==3.0.0a9',
]

# What packages are required for this module to be executed?
REQUIRED = [
    f'aim-ui-custom==3.15.2.post5',
    'aimrecords==0.0.7',
    'aimrocks==0.2.1',
    'cachetools>=4.0.0',
    'click>=7.0',
    'cryptography>=3.0',
    'filelock>=3.3.0',
    'numpy>=1.12.0',
    'protobuf>=3.11.0',
    'psutil>=5.6.7',
    'py3nvml>=0.2.5',
    'RestrictedPython>=5.1',
    'tqdm>=4.20.0',
    'aiofiles>=0.5.0',
    'alembic>=1.4.0',
    'fastapi>=0.65.0,<0.68.0',
    'jinja2>=2.10.0',
    'pytz>=2019.1',
    'SQLAlchemy>=1.4.1',
    'uvicorn>=0.12.0',
    'Pillow>=8.0.0',
    'protobuf >= 3.9.2,<4.0.0',
    # fastapi to support python3.6
    'async-exit-stack>=1.0.0; python_version<"3.7"',
    'async-generator>=1.0; python_version<"3.7"',
    'packaging>=15.0',
    'python-dateutil',
    'requests',
]

if platform.machine() != 'arm64':
    # Temporarily avoid `grpcio` until the issue
    # https://github.com/grpc/grpc/issues/29262 is resolved
    REQUIRED.append('grpcio>=1.42.0')


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = [
        ('rc', None, 'Tag version as a release candidate'),
    ]

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        self.rc = 0

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Cleaning build directory')
        os.system('{} setup.py clean --all'.format(sys.executable))

        self.status('Building Source and Wheel (universal) distribution…')
        os.system(f'{sys.executable} setup.py sdist bdist_wheel --universal')

        # self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        if not self.rc:
            self.status('Pushing git tags…')
            os.system('git tag v{0}'.format(VERSION))
            os.system('git push --tags')

        sys.exit()


INCLUDE_DIRS = [lib_utils.get_include_dir()]
LIB_DIRS = [lib_utils.get_lib_dir()]
LIBS = lib_utils.get_libs()
COMPILE_ARGS = [
    '-std=c++11',
    '-O3',
    '-Wall',
    '-Wextra',
    '-Wconversion',
    '-fno-strict-aliasing',
    '-fno-rtti',
    '-fPIC'
]
CYTHON_SCRITPS = [
    ('aim.storage.hashing.c_hash', 'aim/storage/hashing/c_hash.pyx'),
    ('aim.storage.hashing.hashing', 'aim/storage/hashing/hashing.py'),
    ('aim.storage.hashing', 'aim/storage/hashing/__init__.py'),
    ('aim.storage.encoding.encoding_native', 'aim/storage/encoding/encoding_native.pyx'),
    ('aim.storage.encoding.encoding', 'aim/storage/encoding/encoding.pyx'),
    ('aim.storage.encoding', 'aim/storage/encoding/__init__.py'),
    ('aim.storage.treeutils', 'aim/storage/treeutils.pyx'),
    ('aim.storage.rockscontainer', 'aim/storage/rockscontainer.pyx'),
    ('aim.storage.union', 'aim/storage/union.pyx'),
    ('aim.storage.arrayview', 'aim/storage/arrayview.py'),
    ('aim.storage.treearrayview', 'aim/storage/treearrayview.py'),
    ('aim.storage.treeview', 'aim/storage/treeview.py'),
    ('aim.storage.utils', 'aim/storage/utils.py'),
    ('aim.storage.container', 'aim/storage/container.py'),
    ('aim.storage.containertreeview', 'aim/storage/containertreeview.py'),
    ('aim.storage.inmemorytreeview', 'aim/storage/inmemorytreeview.py'),
    ('aim.storage.prefixview', 'aim/storage/prefixview.py'),
]


def configure_extension(name: str, path: str):
    """Configure an extension and bind with third-party libs"""
    if isinstance(path, str):
        path = [path]
    return Extension(
        name,
        path,
        language='c++',
        include_dirs=INCLUDE_DIRS,
        libraries=LIBS,
        library_dirs=LIB_DIRS,
        extra_compile_args=COMPILE_ARGS,
    )


def cytonize_extensions():
    """Configure and Cythonize all the extensions"""
    extensions = []
    for name, path in CYTHON_SCRITPS:
        extensions.append(configure_extension(name, path))
    return cythonize(extensions, show_all_warnings=True)


# Where the magic happens
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    python_requires=REQUIRES_PYTHON,
    setup_requires=SETUP_REQUIRED,
    install_requires=REQUIRED,
    packages=packages,
    package_data={'aim': migration_files + storage_migration_files + notifier_files + version_files},
    include_package_data=True,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    ext_modules=cytonize_extensions(),
    entry_points={
        'console_scripts': [
            'aim=aim.cli.cli:cli_entry_point',
            'aim-watcher=aim.cli.watcher_cli:cli_entry_point',
        ],
    },
    cmdclass={
        'upload': UploadCommand
    }
)



from copy import deepcopy
from typing import Iterator

from aim.storage.hashing import hash_auto
from aim.storage.types import AimObject, AimObjectKey


class Context:
    __slots__ = ('_context', '_hash')

    def __init__(
        self,
        context: AimObject
    ):
        if context is None:
            context = {}
        self._context = deepcopy(context)
        self._hash = None

    @property
    def idx(self) -> int:
        return hash(self)

    def __repr__(self) -> str:
        return f'<Context#{hash(self)} {self._context}>'

    def to_dict(self) -> AimObject:
        return self._context

    def __iter__(self) -> Iterator[AimObjectKey]:
        return iter(self._context)

    def __len__(self) -> int:
        return len(self._context)

    def __getitem__(self, idx) -> AimObject:
        return self._context[idx]

    def _calc_hash(self) -> int:
        return hash_auto(self._context)

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = self._calc_hash()
        return self._hash

    def __eq__(self, other: 'Context') -> bool:
        if hash(self) != hash(other):
            return False
        return self._context == other._context



from pathlib import Path
import re
from tempfile import TemporaryDirectory

import click
from tqdm import tqdm

from aim import Run, Distribution, Text, Figure
from aim.ext.resource.log import LogLine
from aim.ext.resource.configs import AIM_RESOURCE_METRIC_PREFIX


def parse_wandb_logs(repo_inst, entity, project, run_id):
    try:
        import wandb
    except ImportError:
        click.echo("Could not process wandb logs - failed to import 'wandb' module.", err=True)
        return

    client = wandb.Api()

    if run_id is None:
        # process all runs
        runs = client.runs(entity + "/" + project)
    else:
        try:
            # get the run by run_id
            run = client.run(f"{entity}/{project}/{run_id}")
        except Exception:
            click.echo(f"Could not find run '{entity}/{project}/{run_id}'", err=True)
            return
        runs = (run,)

    for run in tqdm(runs, desc="Converting wandb logs"):
        if not run.config.items():
            continue
        aim_run = Run(
            repo=repo_inst,
            system_tracking_interval=None,
            capture_terminal_logs=False,
            experiment=project
        )
        aim_run['wandb_run_id'] = run.id
        aim_run['wandb_run_name'] = run.name
        aim_run.description = run.notes

        # Convert Rich Artifacts (Table & Molecule & Plotly & Histogram)
        with TemporaryDirectory() as tmpdirname:
            for step_record in tqdm(run.history(pandas=False), desc='convert rich steps'):
                step = step_record['_step']
                for k, v in step_record.items():
                    if not isinstance(v, (list, dict)):
                        continue
                    rs = v if isinstance(v, list) else [v]
                    for r in rs:
                        if r['_type'] == 'histogram':
                            d = Distribution(hist=r['values'], bin_count=len(r['bins']), bin_range=(r['bins'][0], r['bins'][-1]))                                                                                          
                            aim_run.track(d, name=k, step=step)
                        elif r['_type'] == 'plotly-file' or 'plotly.json' in r['path']:
                            path = r['path']
                            run.file(path).download(root=tmpdirname) 
                            from plotly.io import read_json                 
                            figure = read_json(Path(tmpdirname) / path)
                            aim_run.track(Figure(figure), name=k, step=step)   
                        elif r['_type'] == 'table-file':
                            path = r['path']
                            run.file(path).download(root=tmpdirname)
                            with open(Path(tmpdirname) / path) as f:
                                table_content = f.read()
                            aim_run.track(Text(f'data:text/table,{table_content}'), name=k, step=step)  
                        elif r['_type'] == 'molecule-file':
                            path = r['path']
                            run.file(path).download(root=tmpdirname)
                            with open(Path(tmpdirname) / path) as f:
                                molecule_content = f.read()
                            dtype = Path(path).suffix.lstrip('.')
                            aim_run.track(Text(f'data:text/{dtype},{molecule_content}'), name=k, step=step)  
        
        with TemporaryDirectory() as tmpdirname:
            # Collect console output logs
            console_log_filename = 'output.log'
            console_log_file = run.file(console_log_filename)
            try:
                # Even though the file does not exist, a file object will be returned in zero-sized.
                if console_log_file.size:
                    console_log_file.download(root=tmpdirname)
                    with open(Path(tmpdirname) / console_log_filename) as f:
                        [aim_run.track(LogLine(line), name='logs', step=i) for i, line in enumerate(f.readlines())]
            except Exception:
                click.echo("Failed to track console output log.", err=True)

            # TODO: Collect media files, possibly?

        # Collect params & tags
        aim_run['params'] = run.config
        for tag in run.tags:
            aim_run.add_tag(tag)

        from pprint import pprint
        keys = [key for key in run.history(stream='default').keys() if not key.startswith('_')]

        # Collect metrics
        for record in run.scan_history():
            step = record.get('_step')
            epoch = record.get('epoch')
            timestamp = record.get('_timestamp')
            for key in keys:
                value = record.get(key)
                if value is None:
                    continue
                try:
                    tag, name = key.rsplit("/", 1)
                    if "train" in tag:
                        context = {'tag': tag, 'subset': 'train'}
                    elif "val" in tag:
                        context = {'tag': tag, 'subset': 'val'}
                    elif "test" in tag:
                        context = {'tag': tag, 'subset': 'test'}
                    else:
                        context = {'tag': tag}
                except ValueError:
                    name, context = key, {}
                try:
                    if timestamp:
                        aim_run._tracker._track(value, track_time=timestamp, name=name,
                                                step=step, epoch=epoch, context=context)
                    else:
                        aim_run.track(value, name=name, step=step, epoch=epoch, context=context)
                except ValueError:
                    pass
                    #click.echo(f"Type '{type(value).__name__}': artifacts are not supported yet.", err=True)

        # Collect system logs
        # NOTE: In 'system' logs, collecting sampled history cannot be avoided. (default 'samples' == 500)
        # TODO: async history fetching for better performance
        for record in run.history(stream='system', pandas=False, samples=1e3):
            timestamp = record.get('_timestamp')
            for key in record:
                if key.startswith('_'):  # Including '_runtime', '_timestamp', '_wandb'
                    continue

                value = record.get(key)
                if value is None:
                    continue

                name, context = _normalize_system_metric_key(key)
                if name is None:
                    continue

                try:
                    if timestamp:
                        aim_run._tracker._track(value, track_time=timestamp,
                                                name=f'{AIM_RESOURCE_METRIC_PREFIX}{name}', context=context)
                    else:
                        aim_run.track(value, name=f'{AIM_RESOURCE_METRIC_PREFIX}{name}', context=context)
                except ValueError:
                    pass
                    #click.echo(f"Type '{type(value).__name__}': artifacts are not supported yet.", err=True)


def _normalize_system_metric_key(key):
    # Remap names for being categorized as `System` in aim ui: `aim/web/ui/src/config/systemMetrics/systemMetrics.ts``
    # {value is None} means not supported yet
    SYSTEM_METRICS_NAME_MAP = {
        '': {
            'cpu': 'cpu',
            'disk': 'disk_percent',
            'memory': 'memory_percent',
            'network.recv': None,
            'network.sent': None,
        },
        'gpu': {
            'gpu': 'gpu',
            'memory': 'gpu_memory_percent',
            'memoryAllocated': None,
            'powerPercent': None,
            'powerWatts': 'gpu_power_watts',
            'temp': 'gpu_temp',
        },
        'proc': {
            'cpu.threads': None,
            'memory.availableMB': None,
            'memory.percent': 'p_memory_percent',
            'momory.rssMB': None,
        }
    }

    name = re.sub(r'^system\.', '', key)
    gpu_idx_pattern = re.compile(r'^[0-9]+\.')

    # Triage & Remap name for aim ui
    if name.startswith('gpu'):
        name = re.sub(r'^gpu\.', '', name)

        # Cut & paste gpu idx from name to context
        gpu_idx_match = gpu_idx_pattern.search(name)
        if gpu_idx_match:
            gpu_idx_str = gpu_idx_match.group()
            name = name[len(gpu_idx_str):]
            gpu_idx = int(gpu_idx_str.rstrip('.'))
            context = {'gpu': gpu_idx, 'tag': 'system', 'subset': 'gpu'}
        else:
            context = {'gpu': 'no_idx', 'tag': 'system', 'subset': 'gpu'}

        normalized_name = SYSTEM_METRICS_NAME_MAP['gpu'].get(name)

    elif name.startswith('proc'):
        name = re.sub(r'^proc\.', '', name)
        normalized_name = SYSTEM_METRICS_NAME_MAP['proc'].get(name)
        context = {'tag': 'system', 'subset': 'proc'}

    else:
        normalized_name = SYSTEM_METRICS_NAME_MAP[''].get(name)
        context = {'tag': 'system'}

    return normalized_name, context



import aimrocks
from aim.sdk import *
from aim.ext.notebook.notebook import load_ipython_extension
from aim.cli.manager.manager import run_process

from aim.utils.deprecation import python_version_deprecation_check

python_version_deprecation_check()



# Legacy SDK functions
from aim.sdk.legacy.session import Session
from aim.sdk.legacy.flush import flush
from aim.sdk.legacy.init import init
from aim.sdk.legacy.track import track, set_params
from aim.sdk.legacy.select import select_runs, select_metrics

# SDK aliases
from aim.sdk.run import Run
from aim.sdk.repo import Repo
from aim.sdk.training_flow import TrainingFlow

# pre-defined sequences and custom objects
from aim.sdk.objects import Image
from aim.sdk.objects import Audio
from aim.sdk.objects import Figure
from aim.sdk.objects import Distribution
from aim.sdk.objects import Text

from aim.sdk.sequences.image_sequence import Images
from aim.sdk.sequences.audio_sequence import Audios
from aim.sdk.sequences.distribution_sequence import Distributions
from aim.sdk.sequences.figure_sequence import Figures
from aim.sdk.sequences.text_sequence import Texts
from aim.sdk.sequences.metric import Metric



query_0 = 'run.hparams.benchmark == "glue" ' \
          'and run.hparams.dataset == "cola" ' \
          'and metric.context.subset != "train"'
query_1 = 'run.hparams.benchmark == "glue" ' \
          'and run.hparams.dataset == "cola"'
query_2 = 'run.hparams.benchmark == "glue"'
query_3 = 'run.hparams.dataset == "cola" ' \
          'and run.experiment.name != "baseline-warp_4-cola"'


queries = {
    0: query_0,
    1: query_1,
    2: query_2,
    3: query_3,
}



import datetime
import json

from aim.sdk.sequence import Sequence
from aim.storage import treeutils

from typing import Union, Tuple
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pandas import DataFrame


class Metric(Sequence):
    """Class representing series of numeric values."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        # TODO remove 'float64': temporary fix for repos generated with aim < 3.0.7
        return 'float', 'float64', 'int', 'number'

    @classmethod
    def sequence_name(cls) -> str:
        return 'metric'

    def dataframe(
        self,
        include_name: bool = False,
        include_context: bool = False,
        include_run: bool = False,
        only_last: bool = False
    ) -> 'DataFrame':
        """Get metric series as pandas DataFrame

        Args:
             include_name: (:obj:`int`, optional): If true, include metric name in dataframe. False by default.
             include_context: (:obj:`int`, optional): If true, include metric context
                path:value pairs in dataframe. False by default.
             include_run: (:obj:`int`, optional): If true, include run run.hash and run hparams
                path:value pairs in dataframe. False by default.
             only_last: (:obj:`int`, optional): If true return dataframe for only last value, step, timestamp
                and epoch. False by default.
        """
        # Returns dataframe with rows corresponding to iters
        # Columns: `step`, `value`, `time`
        # steps = list(self.steps)
        self.preload()

        if only_last:
            last_step, last_value = self.values.last()
            steps = [last_step] if self.version == 1 else [self.data.steps[last_step]]
            values = [last_value]
            epochs = [self.epochs[last_step]]
            timestamps = [self.timestamps[last_step]]
        else:
            try:
                steps, (values, epochs, timestamps) = self.data.items_list()
            except ValueError:
                steps = []
                values = []
                epochs = []
                timestamps = []
        indices = [i for i, _ in enumerate(steps)]
        timestamps = [datetime.datetime.fromtimestamp(t) for t in timestamps]
        data = {
            'idx': indices,
            'step': steps,
            'value': values,
            'epoch': epochs,
            'time': timestamps
        }

        if include_run:
            data['run.hash'] = [self.run.hash] * len(indices)
            for path, val in treeutils.unfold_tree(self.run[...],
                                                   unfold_array=False,
                                                   depth=3):
                s = 'run'
                for key in path:
                    if isinstance(key, str):
                        s += f'.{key}'
                    else:
                        s += f'[{key}]'

                if isinstance(val, (tuple, list, dict)):
                    val = json.dumps(val)
                data[s] = [val for _ in indices]
        if include_name:
            data['metric.name'] = [self.name for _ in indices]
        if include_context:
            for path, val in treeutils.unfold_tree(self.context.to_dict(),
                                                   unfold_array=False,
                                                   depth=3):
                s = 'metric.context'
                for key in path:
                    if isinstance(key, str):
                        s += f'.{key}'
                    else:
                        s += f'[{key}]'
                # path = '.'.join(path)
                if isinstance(val, (tuple, list)):
                    val = json.dumps(val)
                # df[s] = val
                data[s] = [val for _ in indices]

        import pandas as pd
        df = pd.DataFrame(data)
        return df



from performance_tests.base import StorageTestBase
from performance_tests.utils import get_baseline, write_baseline
from performance_tests.storage.utils import (
    collect_sequence_containers,
    open_containers_for_read
)


class TestContainerOpenExecutionTime(StorageTestBase):
    def test_container_open(self):
        test_name = 'test_container_open'
        containers_to_open = collect_sequence_containers()
        execution_time = open_containers_for_read(containers_to_open)
        baseline = get_baseline(test_name)
        if baseline:
            self.assertInRange(execution_time, baseline)
        else:
            write_baseline(test_name, execution_time)



from aim.storage.types import AimObject


def from_omegaconf_config(obj):
    try:
        from omegaconf import OmegaConf
    except ModuleNotFoundError:
        return

    if OmegaConf.is_config(obj):
        return OmegaConf.to_container(obj, resolve=True)


def convert_to_native_object(
    obj: AimObject,
    *,
    strict: bool = True,
):
    converters = [
        from_omegaconf_config
    ]
    for func in converters:
        _obj = func(obj)
        if _obj is not None:
            obj = _obj
            break
    else:
        if not strict:
            obj = repr(obj)
        else:
            raise TypeError(f'Unhandled non-native value `{obj}` of type `{type(obj)}`.')

    return obj



import os
import logging

from filelock import BaseFileLock, SoftFileLock, UnixFileLock, has_fcntl

from cachetools.func import ttl_cache
from psutil import disk_partitions

from typing import Optional, Union, Dict, Set, Tuple

logger = logging.getLogger(__name__)


class FileSystemInspector:
    """
    This class is used to inspect the file system state and determine if the
    file being created can use `fcntl` / `flock` or should use soft file locks.

    This can be achieved by finding the device owning the parent directory and
    mapping it to the partition and finally checking the filesystem type.

    We prefer using soft file locks *only* if there is a risk of the filesystem
    not being able to handle `fcntl` implementations.
    (e.g. remote filesystems, including NFS<4, BeeGFS, Lustre, etc.)
    """
    GLOBAL_FILE_LOCK_CAPABLE_FILESYSTEMS: Set[str] = {
        # The most popular local filesystems
        'ext3', 'ext4', 'xfs', 'zfs', 'apfs',
        # We include only NFS v4 filesystems here (yet).
        # NFS v3 does not support file locks.
        'nfs4',
    }
    warned_devices: Set[int] = set()

    @classmethod
    def _warn_only_once(
        cls,
        path: str,
        device: int,
        fstype: Optional[str],
    ) -> None:
        """
        Warn only once per device. This is used to avoid spamming the logs with
        the same message.
        """
        if device in cls.warned_devices:
            return
        cls.warned_devices.add(device)

        if fstype is None:
            logger.warning(
                f"Failed to determine filesystem type for {path} "
                f"(device id: {device}). "
                f"Using soft file locks to avoid potential data corruption."
            )
            return

        logger.warning(
            f"The lock file {path} is on a filesystem of type `{fstype}` "
            f"(device id: {device}). "
            f"Using soft file locks to avoid potential data corruption."
        )

    @classmethod
    @ttl_cache(ttl=10)
    def dev_fstype_mapping(cls) -> Dict[int, str]:
        """
        Returns a mapping of device numbers to filesystem types.
        """
        mapping: Dict[int, str] = {}
        for partition in disk_partitions(all=True):
            try:
                device_id = os.lstat(partition.mountpoint).st_dev
            except (OSError, ValueError, AttributeError):
                continue
            mapping[device_id] = partition.fstype
        return mapping

    @classmethod
    def get_fs(cls, path: str) -> Tuple[Optional[int], Optional[str]]:
        """
        Returns the device id and the filesystem type of the file / directory
        at the given path.

        If the path does not exist, `None` is returned.
        """
        try:
            stat = os.stat(path)
        except (OSError, ValueError, AttributeError):
            return None, None

        return stat.st_dev, cls.dev_fstype_mapping().get(stat.st_dev)

    @classmethod
    def needs_soft_lock(cls, path: str) -> bool:
        """
        Returns `True` if the file system of the given path is not capable of
        using `fcntl` locks.

        This simply checks the device that the parent node resides on and
        depending on the filesystem type, returns `True` or `False`.

        For unknown filesystem types, we'll use soft file locks.
        """
        if not has_fcntl:
            # We can't use `fcntl` locks on this platform no matter the
            # underlying filesystem type. The `filelock` package will log a
            # warning if this happens. No need to log it again.
            return True

        dirname = os.path.dirname(path)
        device_id, fstype = cls.get_fs(dirname)

        # We'll move forward with `fcntl` locks for certain filesystems.
        if fstype in cls.GLOBAL_FILE_LOCK_CAPABLE_FILESYSTEMS:
            return False

        # In all other cases, we'll use soft file locks.
        cls._warn_only_once(path, device_id, fstype)
        return True


def AutoFileLock(
    lock_file: Union[str, os.PathLike],
    timeout: float = -1
) -> BaseFileLock:
    """
    Returns a file lock based on the file system type of the given path.

    If the file system of the given path is not capable of using `fcntl` locks,
    a soft file lock is used.

    If the file system type of the given path is failed to be determined,
    a soft file lock is used.


    :param lock_file: path to the file
    :param timeout: default timeout when acquiring the lock. It will be used as
        fallback value in the acquire method, if no timeout value (``None``) is
        given. If you want to disable the timeout, set it to a negative value.
        A timeout of 0 means, that there is exactly one attempt to acquire the
        file lock.
    """
    if not FileSystemInspector.needs_soft_lock(lock_file):
        return UnixFileLock(lock_file, timeout)
    else:
        # Cleaning lock files is not required by `FileLock`. The leftover lock files
        # (potentially from previous versions) could be interpreted as *acquired*
        # locks by `SoftFileLock` causing a deadlock.
        # To prevent this, we add a suffix to the lock file name.
        return SoftFileLock(f'{lock_file}.softlock', timeout)


class DualLock:
    """ Custom lock that uses both UnixLock and SoftFileLock"""
    def __init__(self, lock_path: Union[str, os.PathLike], timeout: float = -1):
        self._lock_path = str(lock_path)
        self._lock = UnixFileLock(self._lock_path, timeout)

        self._soft_lock_path = f'{self._lock_path}.softlock'
        self._soft_lock = SoftFileLock(self._soft_lock_path, timeout)

    def acquire(self):
        self._lock.acquire()
        self._soft_lock.acquire()

    def release(self):
        self._soft_lock.release()
        self._lock.release()


class NoopLock:
    """No-op lock implementation using duck-typing"""
    def __init__(self, *args, **kwargs):
        pass

    def acquire(self, *args, **kwargs):
        pass

    def release(self, *args, **kwargs):
        pass


class RunLock:
    """Interface for locking/releaseing Run locks."""
    def lock(self, force: bool = False) -> None:
        raise NotImplementedError

    def release(self, force: bool = False) -> None:
        raise NotImplementedError



# Copied from
# https://github.com/grpc/grpc/blob/master/src/python/grpcio_health_checking/grpc_health/v1/health.py
# with minimal changes to work within Aim

import collections
import threading

from aim.ext.transport import health_pb2 as health_pb2
from aim.ext.transport import health_pb2_grpc as health_pb2_grpc


# The service name of the health checking servicer.
SERVICE_NAME = health_pb2.DESCRIPTOR.services_by_name['Health'].full_name
# The entry of overall health for the entire server.
OVERALL_HEALTH = ''


class _Watcher:

    def __init__(self):
        self._condition = threading.Condition()
        self._responses = collections.deque()
        self._open = True

    def __iter__(self):
        return self

    def _next(self):
        with self._condition:
            while not self._responses and self._open:
                self._condition.wait()
            if self._responses:
                return self._responses.popleft()
            else:
                raise StopIteration()

    def next(self):
        return self._next()

    def __next__(self):
        return self._next()

    def add(self, response):
        with self._condition:
            self._responses.append(response)
            self._condition.notify()

    def close(self):
        with self._condition:
            self._open = False
            self._condition.notify()


def _watcher_to_send_response_callback_adapter(watcher):

    def send_response_callback(response):
        if response is None:
            watcher.close()
        else:
            watcher.add(response)

    return send_response_callback


class HealthServicer(health_pb2_grpc.HealthServicer):
    """Servicer handling RPCs for service statuses."""

    def __init__(self,
                 experimental_non_blocking=True,
                 experimental_thread_pool=None):
        self._lock = threading.RLock()
        self._server_status = {"": health_pb2.HealthCheckResponse.SERVING}
        self._send_response_callbacks = {}
        self.Watch.__func__.experimental_non_blocking = experimental_non_blocking
        self.Watch.__func__.experimental_thread_pool = experimental_thread_pool
        self._gracefully_shutting_down = False

    def _on_close_callback(self, send_response_callback, service):

        def callback():
            with self._lock:
                self._send_response_callbacks[service].remove(
                    send_response_callback)
            send_response_callback(None)

        return callback

    def Check(self, request, context):
        import grpc
        with self._lock:
            status = self._server_status.get(request.service)
            if status is None:
                context.set_code(grpc.StatusCode.NOT_FOUND)
                return health_pb2.HealthCheckResponse()
            else:
                return health_pb2.HealthCheckResponse(status=status)

    # pylint: disable=arguments-differ
    def Watch(self, request, context, send_response_callback=None):
        blocking_watcher = None
        if send_response_callback is None:
            # The server does not support the experimental_non_blocking
            # parameter. For backwards compatibility, return a blocking response
            # generator.
            blocking_watcher = _Watcher()
            send_response_callback = _watcher_to_send_response_callback_adapter(
                blocking_watcher)
        service = request.service
        with self._lock:
            status = self._server_status.get(service)
            if status is None:
                status = health_pb2.HealthCheckResponse.SERVICE_UNKNOWN  # pylint: disable=no-member
            send_response_callback(
                health_pb2.HealthCheckResponse(status=status))
            if service not in self._send_response_callbacks:
                self._send_response_callbacks[service] = set()
            self._send_response_callbacks[service].add(send_response_callback)
            context.add_callback(
                self._on_close_callback(send_response_callback, service))
        return blocking_watcher

    def set(self, service, status):
        """Sets the status of a service.

        Args:
          service: string, the name of the service.
          status: HealthCheckResponse.status enum value indicating the status of
            the service
        """
        with self._lock:
            if self._gracefully_shutting_down:
                return
            else:
                self._server_status[service] = status
                if service in self._send_response_callbacks:
                    for send_response_callback in self._send_response_callbacks[
                            service]:
                        send_response_callback(
                            health_pb2.HealthCheckResponse(status=status))

    def enter_graceful_shutdown(self):
        """Permanently sets the status of all services to NOT_SERVING.

        This should be invoked when the server is entering a graceful shutdown
        period. After this method is invoked, future attempts to set the status
        of a service will be ignored.

        This is an EXPERIMENTAL API.
        """
        with self._lock:
            if self._gracefully_shutting_down:
                return
            else:
                for service in self._server_status:
                    self.set(service,
                             health_pb2.HealthCheckResponse.NOT_SERVING)  # pylint: disable=no-member
                self._gracefully_shutting_down = True



import collections

import grpc

class _GenericClientInterceptor(grpc.UnaryUnaryClientInterceptor,
                                grpc.UnaryStreamClientInterceptor,
                                grpc.StreamUnaryClientInterceptor,
                                grpc.StreamStreamClientInterceptor):

    def __init__(self, interceptor_function):
        self._fn = interceptor_function
        self.cookie = None

    def process_session_sticky(self, metadata):
        for metadatum in (metadata or []):
            if metadatum.key == 'set-cookie':
                self.cookie = metadatum.value

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            self.cookie,
            client_call_details, iter((request,)), False, False)
        response = continuation(new_details, next(new_request_iterator))
        self.process_session_sticky(response.initial_metadata())
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details,
                               request):
        new_details, new_request_iterator, postprocess = self._fn(
            self.cookie,
            client_call_details, iter((request,)), False, True)
        response_it = continuation(new_details, next(new_request_iterator))
        self.process_session_sticky(response_it.initial_metadata())
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details,
                               request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            self.cookie,
            client_call_details, request_iterator, True, False)
        response = continuation(new_details, new_request_iterator)
        self.process_session_sticky(response.initial_metadata())
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details,
                                request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            self.cookie,
            client_call_details, request_iterator, True, True)
        response_it = continuation(new_details, new_request_iterator)
        self.process_session_sticky(response_it.initial_metadata())
        return postprocess(response_it) if postprocess else response_it


def create(intercept_call):
    return _GenericClientInterceptor(intercept_call)


class _ClientCallDetails(
        collections.namedtuple(
            '_ClientCallDetails',
            ('method', 'timeout', 'metadata', 'credentials')),
        grpc.ClientCallDetails):
    pass


def client_auth_interceptor(value):

    def intercept_call(cookie, client_call_details, request_iterator, request_streaming, response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.append(('authorization', value))
        if cookie:
            metadata.append(('cookie', cookie))  
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return create(intercept_call)


import os
from typing import Optional, Dict
import json

from aim.cli.upgrade._legacy_repo.configs import AIM_OBJECTS_DIR_NAME, AIM_MAP_DIR_NAME, AIM_COMMIT_CONFIG_FILE_NAME
from aim.cli.upgrade._legacy_repo.repo.utils import (
    get_run_objects_dir_path,
    get_run_objects_meta_file_path,
)
from aim.cli.upgrade._legacy_repo.repo.metric import Metric


class Run(object):
    def __init__(self, repo, experiment_name: str, run_hash: str):
        self.repo = repo
        self.experiment_name = experiment_name
        self.run_hash = run_hash
        self._config = None
        self._params = None
        self._metrics = {}  # type: Optional[Dict[str, Metric]]
        self._tmp_all_metrics = None  # type: Optional[Dict[str, Metric]]
        self._storage = None

    def __repr__(self):
        return '<{e}/{h}: {m}>'.format(e=self.experiment_name, h=self.run_hash,
                                       m=list(self._metrics.values()))

    def __hash__(self):
        return hash((self.repo.path, self.experiment_name, self.run_hash))

    def __eq__(self, other):
        return hash(self) == hash(other)

    @property
    def params(self) -> dict:
        if self._params is None:
            self._params = self._load_params()
        return self._params

    @property
    def config(self) -> dict:
        if self._config is None:
            self._config = self._load_config()
        return self._config

    @property
    def storage(self):
        return self._storage

    def open_storage(self):
        if self._storage is None:
            storage_path = get_run_objects_dir_path(self.repo.path,
                                                    self.experiment_name,
                                                    self.run_hash)
            self._storage = self.repo.get_records_storage(storage_path, 'r')

    def close_storage(self):
        if self._storage is not None:
            self._storage.close()

    def get_all_metrics(self) -> Dict[str, Metric]:
        if self._tmp_all_metrics is not None:
            return self._tmp_all_metrics

        meta_file_path = get_run_objects_meta_file_path(self.repo.path,
                                                        self.experiment_name,
                                                        self.run_hash)
        metrics = {}
        try:
            with open(meta_file_path, 'r+') as meta_file:
                artifacts = json.loads(meta_file.read().strip())
                # Filter only metrics
                for artifact in artifacts.values():
                    if artifact['type'] == 'metrics':
                        metric = Metric(self.repo,
                                        self,
                                        artifact['name'],
                                        artifact.get('context'))
                        metrics[artifact['name']] = metric
        except:
            pass

        self._tmp_all_metrics = metrics

        return metrics

    def _load_params(self) -> dict:
        params_file_path = os.path.join(self.repo.path,
                                        self.experiment_name,
                                        self.run_hash,
                                        AIM_OBJECTS_DIR_NAME,
                                        AIM_MAP_DIR_NAME,
                                        'dictionary.log')
        try:
            with open(params_file_path, 'r+') as params_file:
                params = json.loads(params_file.read().strip())
        except:
            params = {}
        return params

    def _load_config(self) -> dict:
        config_file_path = os.path.join(self.repo.path,
                                        self.experiment_name,
                                        self.run_hash,
                                        AIM_COMMIT_CONFIG_FILE_NAME)
        try:
            with open(config_file_path, 'r+') as config_file_path:
                config = json.loads(config_file_path.read().strip())
        except:
            config = {}
        return config



import pytest

from tests.base import TestBase
from tests.utils import is_package_installed
from unittest.mock import patch


class TestDVCIntegration(TestBase):

    @patch("aim.sdk.objects.plugins.dvc_metadata.Repo")
    @pytest.mark.skipif(not is_package_installed('dvc'), reason="'dvc' is not installed. skipping.")
    def test_dvc_files_as_run_param(self, mRepo):
        from aim.sdk.objects.plugins.dvc_metadata import DvcData
        from aim.sdk import Run

        tracked_files = (
            {'path': 'test_file_1.txt'},
            {'path': 'test_file_2.txt'},
            {'path': 'test_file_3.txt'}
        )
        files = [e['path'] for e in tracked_files]
        mRepo.ls.return_value = tracked_files

        run = Run(system_tracking_interval=None)
        run['dvc'] = DvcData()

        # get dataset metadata
        dvc_object = run['dvc']
        dvc_dict = run.get('dvc', resolve_objects=True)

        self.assertTrue(isinstance(dvc_object, DvcData))
        self.assertTrue(isinstance(dvc_dict, dict))
        self.assertEqual('dvc', dvc_dict['dataset']['source'])
        self.assertEqual(files, dvc_dict['dataset']['tracked_files'])



import os
import click

from aim.cli.utils import set_log_level
from aim.sdk.repo import Repo, RepoStatus
from aim.sdk.utils import clean_repo_path
from aim.ext.transport.config import AIM_SERVER_DEFAULT_HOST, AIM_SERVER_DEFAULT_PORT, AIM_SERVER_MOUNTED_REPO_PATH
from aim.ext.transport.server import run_router


@click.command()
@click.option('-h', '--host', default=AIM_SERVER_DEFAULT_HOST, type=str)
@click.option('-p', '--port', default=AIM_SERVER_DEFAULT_PORT, type=int)
@click.option('-w', '--workers', default=1, type=int)
@click.option('--repo', required=False, type=click.Path(exists=True,
                                                        file_okay=False,
                                                        dir_okay=True,
                                                        writable=True))
@click.option('--ssl-keyfile', required=False, type=click.Path(exists=True,
                                                               file_okay=True,
                                                               dir_okay=False,
                                                               readable=True))
@click.option('--ssl-certfile', required=False, type=click.Path(exists=True,
                                                                file_okay=True,
                                                                dir_okay=False,
                                                                readable=True))
@click.option('--log-level', required=False, default='', type=str)
def server(host, port, workers,
           repo, ssl_keyfile, ssl_certfile,
           log_level):
    # TODO [MV, AT] remove code duplication with aim up cmd implementation
    if log_level:
        set_log_level(log_level)

    repo_path = clean_repo_path(repo) or Repo.default_repo_path()
    repo_status = Repo.check_repo_status(repo_path)
    if repo_status == RepoStatus.MISSING:
        init_repo = click.confirm(f'\'{repo_path}\' is not a valid Aim repository. Do you want to initialize it?')

        if not init_repo:
            click.echo('To initialize repo please run the following command:')
            click.secho('aim init', fg='yellow')
            return
        repo_inst = Repo.from_path(repo_path, init=True)
    elif repo_status == RepoStatus.UPDATE_REQUIRED:
        upgrade_repo = click.confirm(f'\'{repo_path}\' requires upgrade. Do you want to run upgrade automatically?')
        if upgrade_repo:
            from aim.cli.upgrade.utils import convert_2to3
            repo_inst = convert_2to3(repo_path, drop_existing=False, skip_failed_runs=False, skip_checks=False)
        else:
            click.echo('To upgrade repo please run the following command:')
            click.secho(f'aim upgrade --repo {repo_path} 2to3', fg='yellow')
            return
    else:
        repo_inst = Repo.from_path(repo_path)

    os.environ[AIM_SERVER_MOUNTED_REPO_PATH] = repo_inst.path

    click.echo(
        click.style('Running Aim Server on repo `{}`'.format(repo_inst),
                    fg='yellow'))
    click.echo('Server is mounted on {}:{}'.format(host, port), err=True)
    click.echo('Press Ctrl+C to exit')

    try:
        run_router(host, port, workers, ssl_keyfile, ssl_certfile)
    except Exception:
        click.echo('Failed to run Aim Tracking Server. '
                   'Please see the logs for details.')
        return



from copy import deepcopy

import aimrocks.lib_rocksdb as interfaces  # noqa: F401


class KeysIterator:
    def __init__(self, items_iterator):
        self.it = items_iterator

    def __iter__(self):
        return self

    def __next__(self):
        key, _ = next(self.it)
        return key


class ValuesIterator:
    def __init__(self, items_iterator):
        self.it = items_iterator

    def __iter__(self):
        return self

    def __next__(self):
        _, value = next(self.it)
        return value


class ArrayFlagType:
    def __repr__(self):
        return "<ArrayFlag>"

    def __copy__(self) -> 'ArrayFlagType':
        return self

    def __deepcopy__(self, memo) -> 'ArrayFlagType':
        return self


ArrayFlag = ArrayFlagType()


class ObjectFlagType:
    def __repr__(self):
        return "<ObjectFlag>"

    def __copy__(self) -> 'ObjectFlagType':
        return self

    def __deepcopy__(self, memo) -> 'ObjectFlagType':
        return self


ObjectFlag = ObjectFlagType()


class CustomObjectFlagType:
    def __init__(self, aim_name: str):
        self.aim_name = aim_name

    def __repr__(self):
        return f"<CustomObjectFlag type={self.aim_name}>"


class BLOB:
    def __init__(
        self,
        data=None,
        loader_fn=None
    ):
        self.data = data
        self.loader_fn = loader_fn

    def __bytes__(self):
        return bytes(self.load())

    def __len__(self):
        return len(bytes(self))

    def load(self):
        if self.data is None:
            assert self.loader_fn is not None
            self.data = self.loader_fn()
            self.loader_fn = None
        return self.data

    def __deepcopy__(self, memo):
        data = self.load()
        instance = self.__class__(data=deepcopy(data, memo=memo))
        memo[id(self)] = instance
        return instance

    def transform(self, transform_fn):
        if self.data is not None:
            return self.__class__(transform_fn(self.data))

        def loader():
            return transform_fn(self.load())

        return self.__class__(loader_fn=loader)



from tests.base import PrefilledDataApiTestBase


class TestDashboardAppsApi(PrefilledDataApiTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        for _ in range(5):
            cls.client.post('/api/apps/', json={'type': 'metric_explorer', 'state': {}})

    def test_list_apps_api(self):
        response = self.client.get('/api/apps/')
        self.assertEqual(5, len(response.json()))

    def test_get_app_api(self):
        list_response = self.client.get('/api/apps/')
        app_id = list_response.json()[0]['id']
        response = self.client.get(f'/api/apps/{app_id}/')
        data = response.json()
        self.assertEqual(200, response.status_code)
        self.assertEqual(app_id, data['id'])
        self.assertEqual('metric_explorer', data['type'])
        self.assertEqual({}, data['state'])

    def test_update_app_api(self):
        list_response = self.client.get('/api/apps/')
        app_id = list_response.json()[0]['id']
        response = self.client.put(f'/api/apps/{app_id}/', json={'type': 'params_explorer', 'state': {'foo': 'bar'}})
        self.assertEqual(response.status_code, 200)
        get_response = self.client.get(f'/api/apps/{app_id}/')
        data = get_response.json()
        self.assertEqual('params_explorer', data['type'])
        self.assertEqual('bar', data['state']['foo'])

    def test_delete_app_api(self):
        list_response = self.client.get('/api/apps/')
        app_id = list_response.json()[0]['id']
        response = self.client.delete(f'/api/apps/{app_id}/')
        self.assertEqual(200, response.status_code)
        get_response = self.client.get(f'/api/apps/{app_id}/')
        self.assertEqual(404, get_response.status_code)

    def test_create_app_api(self):
        app_data = {'state': {'test_key1': 'test_value', 'test_key2': [1, 2, 3]},
                    'type': 'metric_explorer'}
        response = self.client.post('/api/apps/', json=app_data)
        self.assertEqual(201, response.status_code)
        app_id = response.json()['id']
        get_response = self.client.get(f'/api/apps/{app_id}/')
        data = get_response.json()
        self.assertEqual(app_id, data['id'])
        self.assertDictEqual(app_data['state'], data['state'])
        self.assertEqual(app_data['type'], data['type'])


class TestDashboardsApi(PrefilledDataApiTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        app_data = {"type": "metric_explorer", "state": {}}
        response = cls.client.post('/api/apps/', json=app_data)
        cls.app_id = response.json()['id']
        for i in range(5):
            cls.client.post('/api/dashboards/', json={'name': f'dashboard_{i}'})

    def test_list_dashboards_api(self):
        response = self.client.get('/api/dashboards/')
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(5, len(data))

    def test_get_dashboard_api(self):
        list_response = self.client.get('/api/dashboards/')
        list_data = list_response.json()
        dashboard_id = list_data[0]['id']
        dashboard_name = list_data[0]['name']
        response = self.client.get(f'/api/dashboards/{dashboard_id}/')
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertEqual(dashboard_id, data['id'])
        self.assertEqual(dashboard_name, data['name'])

    def test_update_dashboard_api(self):
        list_response = self.client.get('/api/dashboards/')
        list_data = list_response.json()
        dashboard_id = list_data[0]['id']
        response = self.client.put(f'/api/dashboards/{dashboard_id}', json={'description': 'new description'})
        self.assertEqual(200, response.status_code)

        get_response = self.client.get(f'/api/dashboards/{dashboard_id}')
        data = get_response.json()
        self.assertEqual('new description', data['description'])

    def test_create_dashboard_api(self):
        response = self.client.post('/api/dashboards/', json={'name': 'awesome',
                                                              'description': 'pretty awesome dashboard',
                                                              'app_id': self.app_id})
        self.assertEqual(201, response.status_code)
        dashboard_id = response.json()['id']
        get_response = self.client.get(f'/api/dashboards/{dashboard_id}/')
        data = get_response.json()
        self.assertEqual(dashboard_id, data['id'])
        self.assertEqual('awesome', data['name'])
        self.assertEqual('pretty awesome dashboard', data['description'])
        self.assertEqual(self.app_id, data['app_id'])

    def test_delete_dashboard_api(self):
        list_response = self.client.get('/api/dashboards/')
        list_data = list_response.json()
        dashboard_id = list_data[0]['id']
        response = self.client.delete(f'/api/dashboards/{dashboard_id}/')
        self.assertEqual(200, response.status_code)
        get_response = self.client.get(f'/api/dashboards/{dashboard_id}/')
        self.assertEqual(404, get_response.status_code)



from typing import Generic, Union, Tuple, TypeVar, Dict, Iterator, Any, List
from itertools import islice
import numpy as np
import logging

from aim.sdk.tracker import STEP_HASH_FUNCTIONS
from aim.storage.treeview import TreeView
from aim.storage.arrayview import ArrayView
from aim.storage.context import Context
from aim.storage.hashing import hash_auto

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from aim.sdk.run import Run


logger = logging.getLogger(__name__)


T = TypeVar('T')


class SequenceData:
    def __init__(self, series_tree, version: int, columns: List[Tuple[str, str]]):
        if len(columns) == 0:
            raise ValueError('Cannot create SequenceData. Please specify at least one column.')

        self.series_tree = series_tree
        self.version = version
        self.columns = columns
        self.arrays: Tuple[ArrayView] = tuple(self._get_array(col, dtype) for col, dtype in columns)

        self._dtype_map = {col: dtype for col, dtype in columns}
        self._step_hash_fn = STEP_HASH_FUNCTIONS[self.version]

    def step_hash(self, step):
        return self._step_hash_fn(step)

    def _checked_columns(self, columns: Union[str, List[str]]) -> List[Tuple[str, str]]:
        if isinstance(columns, str):
            columns = [columns]
        return [(col, self._dtype_map[col]) for col in columns if (col in self._dtype_map)]

    def _get_array(self, column: str, dtype: str = None) -> ArrayView:
        return self.series_tree.array(column, dtype=dtype)

    def view(self, columns: Union[str, Tuple[str]]):
        raise NotImplementedError

    def range(self, start, stop) -> 'SequenceData':
        raise NotImplementedError

    def sample(self, k) -> 'SequenceData':
        raise NotImplementedError

    def __iter__(self) -> Iterator[Tuple[int, Any]]:
        yield from self.items()

    def items(self) -> Iterator[Tuple[int, Any]]:
        steps, vals_list = self.items_list()
        yield from zip(steps, zip(*vals_list))

    def values(self) -> Iterator[Any]:
        yield from zip(self.values_list())

    def indices(self) -> Iterator[int]:
        yield from self.indices_list()

    def items_list(self) -> Tuple[List[int], List[Any]]:
        raise NotImplementedError

    def values_list(self) -> List[Any]:
        # default implementation
        return self.items_list()[1]

    def indices_list(self) -> List[int]:
        # default implementation
        return self.items_list()[0]

    def numpy(self) -> Tuple[np.ndarray, List[np.ndarray]]:
        # default implementation
        steps, vals_list = self.items_list()
        numpy_list = []
        for col_idx, vals in enumerate(vals_list):
            numpy_list.append(np.array(vals, dtype=self.arrays[col_idx].dtype))
        return np.array(steps, np.intp), numpy_list


class SequenceV1Data(SequenceData):
    def __init__(
            self,
            series_tree, *,
            columns: List[Tuple[str, str]],
            step_range: Tuple[int, int] = None,
            n_items: int = -1
    ):
        super().__init__(series_tree, version=1, columns=columns)
        self.step_range = step_range
        self.n_items = n_items

    def view(self, columns: List[str]) -> 'SequenceData':
        return SequenceV1Data(
            self.series_tree, columns=self._checked_columns(columns), step_range=self.step_range, n_items=self.n_items)

    def range(self, start, stop) -> 'SequenceData':
        return SequenceV1Data(self.series_tree, columns=self.columns, step_range=(start, stop), n_items=self.n_items)

    def sample(self, k) -> 'SequenceData':
        return SequenceV1Data(self.series_tree, columns=self.columns, step_range=self.step_range, n_items=k)

    def items_list(self) -> Tuple[List[int], List[Any]]:
        iters = self._get_iters()
        columns = [[] for _ in iters]
        steps = []

        if self.step_range is not None:
            start, stop = self.step_range
            if stop <= start or start < 0 or stop < 0:
                return
            for idx, val in iters[0]:
                if idx < start:
                    for it in iters[1:]:
                        next(it)
                    continue
                if idx >= stop:
                    break
                steps.append(idx)
                columns[0].append(val)
                for it_index, it in enumerate(iters[1:]):
                    columns[it_index + 1].append(next(it))
        else:
            for idx, val in iters[0]:
                steps.append(idx)
                columns[0].append(val)
                for it_index, it in enumerate(iters[1:]):
                    columns[it_index + 1].append(next(it))
        length = len(steps)
        step = (length // self.n_items or 1) if self.n_items > 0 else 1
        if step != 1:
            slice_ = slice(0, length, step)
            last_step_needed = (length - 1) % step != 0
            steps = steps[slice_] + [steps[-1]] if last_step_needed else steps[slice_]
            for i, v in enumerate(columns):
                columns[i] = v[slice_] + [v[-1]] if last_step_needed else v[slice_]

        return steps, columns

    def _get_iters(self) -> List[Iterator[Any]]:
        iters = [self.arrays[0].items()]
        for arr in self.arrays[1:]:
            iters.append(arr.values())
        return iters


class SequenceV2Data(SequenceData):
    def __init__(
            self,
            meta_tree, series_tree, *,
            columns: List[Tuple[str, str]],
            n_items: int = -1
    ):
        super().__init__(series_tree, version=2, columns=columns)
        # `SequenceV2Data` has access to both metadata and series data
        # trees that are not necessarily based on the same physical storage.
        # Therefore, we don't have a consistency guarantee between the two.
        # The implemented methods should tolerate this.
        self.meta_tree = meta_tree
        self.n_items = n_items
        self.steps: ArrayView = self._get_array('step')

    def view(self, columns: List[str]) -> 'SequenceData':
        return SequenceV2Data(
            self.meta_tree, self.series_tree, columns=self._checked_columns(columns), n_items=self.n_items)

    def range(self, start, stop) -> 'SequenceData':
        raise ValueError('Range selection cannot be applied to data stored with reservoir sampling.')

    def sample(self, k) -> 'SequenceData':
        return SequenceV2Data(self.meta_tree, self.series_tree, columns=self.columns, n_items=k)

    def items_list(self) -> Tuple[List[int], List[Any]]:
        steps, values = self.numpy()
        return steps.tolist(), [v.tolist() for v in values]

    def numpy(self) -> Tuple[np.ndarray, List[np.ndarray]]:
        if self.n_items == -1:  # select all
            last_step = None
            steps = np.fromiter(self.steps.values(), np.intp)
            columns = [np.fromiter(arr.values(), arr.dtype) for arr in self.arrays]
        else:
            last_step = self.meta_tree['last_step']
            steps = np.fromiter(islice(self.steps.values(), self.n_items), np.intp)
            columns = [np.fromiter(islice(arr.values(), self.n_items), arr.dtype) for arr in self.arrays]

        # sort all columns by step
        sort_indices = steps.argsort()
        columns = [arr[sort_indices] for arr in columns]
        steps = steps[sort_indices]
        if last_step is not None and last_step != steps[-1]:
            step_hash = self.step_hash(last_step)
            # The `last_step` is provided by the meta tree which may potentially
            # be out of sync with the series tree.
            # If such case occurs, we fall back to the series tree for the last step.
            last_steps = []
            try:
                for i in range(len(columns)):
                    last_steps.append(self.arrays[i][step_hash])
            except KeyError:
                logger.debug('Last step not found in reservoir.')
            else:
                # Only if all the last steps are found, we use them.
                for i in range(len(columns)):
                    columns[i][-1] = last_steps[i]
                steps[-1] = last_step

        return steps, columns


class Sequence(Generic[T]):
    """Class representing single series of tracked value.

    Objects series can be retrieved as Sequence regardless the object's type,
    but subclasses of Sequence might provide additional functionality.
    Provides interface to access tracked values, steps, timestamps and epochs.
    Values, epochs and timestamps are accessed via :obj:`aim.storage.arrayview.ArrayView` interface.
    """

    registry: Dict[str, 'Sequence'] = dict()
    collections_allowed = False

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        subclass_typename = cls.sequence_name()
        if subclass_typename is not None:  # check for intermediate helper classes
            cls.registry[subclass_typename] = cls

    def __init__(
        self,
        name: str,
        context: Context,  # TODO ?dict
        run: 'Run'
    ):
        self._hash: int = None
        self._version: int = None
        self._meta_tree: TreeView = run.meta_run_tree.subtree(('traces', context.idx, name))
        self._series_tree: TreeView = None
        self._columns = [('val', None), ('epoch', 'float64'), ('time', 'float64')]
        self._data: SequenceData = None  # use data property

        self.name = name
        self.context = context
        self.run = run

    def __repr__(self) -> str:
        return f'<Sequence#{hash(self)} name=`{self.name}` context=`{self.context}` run=`{self.run}`>'

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        """classmethod to get allowed object types for particular sequence

        For example, numeric sequences a.k.a. Metric allow float and integer numbers.
        The base Sequence allows any value, and to indicate that, `allowed_dtypes` returns '*'.
        """
        return '*'

    @classmethod
    def sequence_name(cls) -> str:
        """classmethod to get retrieve sequence's registered name"""
        ...

    def _calc_hash(self):
        return hash_auto(
            (self.name,
             hash(self.context),
             hash(self.run))
        )

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = self._calc_hash()
        return self._hash

    @property
    def series_tree(self) -> TreeView:
        if self._series_tree is None:
            self._series_tree = self.run.series_run_trees[self.version].subtree((self.context.idx, self.name))
        return self._series_tree

    @property
    def data(self) -> SequenceData:
        if self._data is None:
            if self.version == 1:
                self._data = SequenceV1Data(self.series_tree, columns=self._columns)
            else:
                self._data = SequenceV2Data(self._meta_tree, self.series_tree, columns=self._columns)
        return self._data

    @property
    def version(self):
        if self._version is None:
            self._version = self._meta_tree.get('version', 1)
        return self._version

    @property
    def values(self) -> ArrayView:
        """Tracked values array as :obj:`ArrayView`.

            :getter: Returns values ArrayView.
        """
        return self.data._get_array('val')

    @property
    def epochs(self) -> ArrayView:
        """Tracked epochs array as :obj:`ArrayView`.

            :getter: Returns epochs ArrayView.
        """
        return self.data._get_array('epoch', dtype='float64')

    @property
    def timestamps(self) -> ArrayView:
        """Tracked timestamps array as :obj:`ArrayView`.

            :getter: Returns timestamps ArrayView.
        """
        return self.data._get_array('time', dtype='float64')

    def __bool__(self) -> bool:
        try:
            return bool(self.values)
        except ValueError:
            return False

    def __len__(self) -> int:
        return len(self.values)

    def preload(self):
        self.series_tree.preload()

    def first_step(self):
        """Get sequence tracked first step.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree['first_step']

    def last_step(self):
        """Get sequence tracked last step.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree['last_step']


class MediaSequenceBase(Sequence):
    """Helper class for media sequence types."""

    collections_allowed = True

    def record_length(self):
        """Get tracked records longest list length or `None` if Text objects are tracked.

        Required to implement ranged and sliced data fetching.
        """
        return self._meta_tree.get('record_max_length', None)



from parameterized import parameterized

from tests.base import PrefilledDataTestBase
from tests.utils import full_class_name

from aim.sdk.types import QueryReportMode
from aim.storage.query import syntax_error_check


class TestQuery(PrefilledDataTestBase):

    def test_query_metrics(self):
        q = self.isolated_query_patch('run.hparams.batch_size == None and metric.context.is_training == True')

        trace_count = 0
        for trc in self.repo.query_metrics(query=q, report_mode=QueryReportMode.DISABLED):
            self.assertIsNone(trc.run['hparams']['batch_size'])
            self.assertTrue(trc.context['is_training'])
            trace_count += 1
        self.assertEqual(20, trace_count)

    def test_query_metrics_empty_query(self):
        q = self.isolated_query_patch()
        trcs = self.repo.query_metrics(query=q, report_mode=QueryReportMode.DISABLED)
        trace_count = sum(1 for _ in trcs)
        self.assertEqual(40, trace_count)

    def test_query_runs(self):
        q = self.isolated_query_patch('run.hparams.lr < 0.01 and run.run_index >= 5')

        run_count = 0
        for run_trace_collection in self.repo.query_runs(query=q, report_mode=QueryReportMode.DISABLED).iter_runs():
            run = run_trace_collection.run
            self.assertLess(run[('hparams', 'lr')], 0.01)
            self.assertGreaterEqual(run['run_index'], 5)
            run_count += 1
        self.assertEqual(5, run_count)

    def test_query_runs_empty_query(self):
        q = self.isolated_query_patch()
        runs = self.repo.query_runs(query=q, report_mode=QueryReportMode.DISABLED).iter_runs()
        run_count = sum(1 for _ in runs)
        self.assertEqual(10, run_count)

    def test_query_run_structured_params(self):
        q = self.isolated_query_patch('run.name == "Run # 2"')
        run_count = 0
        for run_trace_collection in self.repo.query_runs(query=q, report_mode=QueryReportMode.DISABLED).iter_runs():
            run = run_trace_collection.run
            self.assertEqual("Run # 2", run.name)
            run_count += 1
        self.assertEqual(1, run_count)

    @parameterized.expand([
        ('run + context', 'run.hparams.batch_size == None and metric.context.is_training == True'),
        ('run multiple filters', 'run.hparams.batch_size == None and run.hparams.lr == 0.01'),
        ('context only', 'metric.context["is_training"] == False'),
        ('context + metric name', 'metric.context.is_training == False and metric.name == "loss"'),
        ('__getitem__ interface with tuple', 'run["hparams","lr"] == 0.01'),
        ('__getitem__ interface chaining', 'run["hparams"]["lr"] == 0.01'),
        ('mixed interface', 'run["hparams"].lr == 0.01'),
    ])
    def test_query_execution(self, name, q):
        # crash/no-crash test
        # execute query and iterate over result
        q = self.isolated_query_patch(q)
        for _ in self.repo.query_metrics(q, report_mode=QueryReportMode.DISABLED):
            continue
        for _ in self.repo.query_runs(q, report_mode=QueryReportMode.DISABLED).iter_runs():
            continue

    def test_invalid_query(self):
        q = self.isolated_query_patch('invalid_varialble.get("hparams", "batch_size") == None')
        with self.assertRaises(Exception):
            next(iter(self.repo.query_metrics(q, report_mode=QueryReportMode.DISABLED)))

    def test_query_raise_syntax_error(self):
        q = self.isolated_query_patch('run.hash == "x" && metric.name == "y"')
        with self.assertRaises(SyntaxError):
            next(iter(self.repo.query_metrics(q, report_mode=QueryReportMode.DISABLED)))

        with self.assertRaises(SyntaxError):
            next(iter(self.repo.query_runs(q, report_mode=QueryReportMode.DISABLED)))

    def test_syntax_error_handling(self):
        q = 'run.hash == "x" && metric.name == "y"'
        try:
            syntax_error_check(q)
        except SyntaxError as se:
            self.assertEqual(19, se.offset)  # count for added ()
            self.assertEqual(1, se.lineno)
        else:
            self.fail('SyntaxError is not raised')


class TestQueryDefaultExpression(PrefilledDataTestBase):
    def setUp(self):
        self.run = next((run for run in self.repo.iter_runs() if run.get('testcase') == full_class_name(self.__class__)))
        self.run.archived = True
        self.run_hash = self.run.hash

    def tearDown(self):
        self.run.archived = False

    def test_default_query_run_results(self):
        run_hashes = [run.run.hash for run in self.repo.query_runs(report_mode=QueryReportMode.DISABLED).iter_runs()]
        self.assertNotIn(self.run_hash, run_hashes)

    def test_default_query_metric_results(self):
        run_hashes = [metric.run.hash for metric in self.repo.query_metrics(report_mode=QueryReportMode.DISABLED)]
        self.assertNotIn(self.run_hash, run_hashes)

    def test_query_with_archived_expression_run_results(self):
        q = self.isolated_query_patch('run.archived == True')
        run_hashes = []
        for run in self.repo.query_runs(query=q, report_mode=QueryReportMode.DISABLED).iter_runs():
            run_hashes.append(run.run.hash)
            self.assertTrue(run.run.archived)
        self.assertIn(self.run_hash, run_hashes)

    def test_query_without_archived_expression_metric_results(self):
        q = self.isolated_query_patch('metric.name == "accuracy"')
        run_hashes = []
        for metric in self.repo.query_metrics(query=q, report_mode=QueryReportMode.DISABLED):
            run_hashes.append(metric.run.hash)
            self.assertFalse(metric.run.archived)
        self.assertNotIn(self.run_hash, run_hashes)



import os
import shutil
import tarfile
import pathlib
import re
import uuid
from contextlib import contextmanager
from typing import Union, Any, Tuple, Optional, Callable

from aim.sdk.configs import get_aim_repo_name

from aim.storage.object import CustomObject


def search_aim_repo(path) -> Tuple[Any, bool]:
    found = False
    path = os.path.abspath(path)
    while path:
        repo_path = os.path.join(path, get_aim_repo_name())
        if os.path.exists(repo_path) and os.path.isdir(repo_path):
            found = True
            return path, found
        if path == '/':
            return None, found
        path = os.path.dirname(path)


def generate_run_hash(hash_length=24) -> str:
    return uuid.uuid4().hex[:hash_length]


def clean_repo_path(repo_path: Union[str, pathlib.Path]) -> str:
    if isinstance(repo_path, pathlib.Path):
        repo_path = str(repo_path)

    if not isinstance(repo_path, str) or not repo_path:
        return ''

    repo_path = repo_path.strip().rstrip('/')

    if isinstance(repo_path, pathlib.Path):
        repo_path = str(repo_path)
    if repo_path == '.':
        return os.getcwd()
    if repo_path == '~':
        return os.path.expanduser('~')

    if repo_path.endswith(get_aim_repo_name()):
        repo_path = repo_path[:-len(get_aim_repo_name())]
    if repo_path.startswith('~'):
        repo_path = os.path.expanduser('~') + repo_path[1:]

    return os.path.abspath(repo_path)


def get_object_typename(obj) -> str:
    if isinstance(obj, float):
        return 'float'
    if isinstance(obj, (int, bool)):
        return 'int'
    if isinstance(obj, str):
        return 'str'
    if isinstance(obj, bytes):
        return 'bytes'
    if isinstance(obj, dict):
        return 'object'
    if isinstance(obj, (tuple, list)):
        if len(obj) == 0:
            # element type is unknown yet.
            return 'list'
        element_typename = get_object_typename(obj[0])
        return f'list({element_typename})'
    if isinstance(obj, CustomObject):
        return obj.get_typename()
    return 'unknown'


any_list_regex = re.compile(r'list\([A-Za-z]{1}[A-Za-z0-9.]*\)')


def check_types_compatibility(
        dtype: str,
        base_dtype: str,
        update_base_dtype_fn: Optional[Callable[[str, str], None]] = None) -> bool:
    if dtype == base_dtype:
        return True
    if base_dtype == 'number' and dtype in {'int', 'float'}:
        return True
    if {dtype, base_dtype} == {'int', 'float'}:
        if update_base_dtype_fn is not None:
            update_base_dtype_fn(base_dtype, 'number')
        return True
    if base_dtype == 'list' and any_list_regex.match(dtype):
        if update_base_dtype_fn is not None:
            update_base_dtype_fn(base_dtype, dtype)
        return True
    if dtype == 'list' and any_list_regex.match(base_dtype):
        return True
    return False


@contextmanager
def work_directory(path: str):
    curr = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(curr)


def backup_run(run) -> str:
    repo_path = run.repo.path
    backups_dir = os.path.join(repo_path, 'bcp')
    if not os.path.exists(backups_dir):
        os.mkdir(backups_dir)

    run_bcp_file = f'bcp/{run.hash}'
    with work_directory(repo_path):
        with tarfile.open(run_bcp_file, 'w:gz') as tar:
            for part in ('meta', 'seqs'):
                tar.add(os.path.join(part, 'chunks', run.hash))
    return run_bcp_file


def restore_run_backup(repo, run_hash):
    repo_path = repo.path
    backups_dir = os.path.join(repo_path, 'bcp')

    assert os.path.exists(backups_dir)
    with work_directory(repo_path):
        run_bcp_file = f'bcp/{run_hash}'
        shutil.rmtree(f'meta/chunks/{run_hash}', ignore_errors=True)
        shutil.rmtree(f'seqs/chunks/{run_hash}', ignore_errors=True)
        with tarfile.open(run_bcp_file, 'r:gz') as tar:
            tar.extractall()
        progress_path = pathlib.Path('meta') / 'progress' / run_hash
        progress_path.touch(exist_ok=True)



from aim.cli.upgrade._legacy_repo.repo.repo import AimRepo
from aim.cli.upgrade._legacy_repo.repo.run import Run
from aim.cli.upgrade._legacy_repo.metric_artifact import deserialize_pb
from aim.cli.upgrade._legacy_repo.configs import AIM_MAP_METRICS_KEYWORD



from typing import Union, Tuple

from aim.sdk.sequence import MediaSequenceBase
from aim.sdk.objects import Audio


class Audios(MediaSequenceBase):
    """Class representing series of Audio objects or Audio lists."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        typename = Audio.get_typename()
        return typename, f'list({typename})'

    @classmethod
    def sequence_name(cls) -> str:
        return 'audios'



from aim.storage.container import Container, ContainerItemsIterator, ContainerKey, ContainerValue
from aim.storage.containertreeview import ContainerTreeView

from typing import Iterator, Tuple


class PrefixView(Container):
    """
    A mutable view to a :obj:`Container` given by a key prefix.

    Args:
        prefix (:obj:`bytes`): the prefix that defines the key range of the
            view-container. The resulting container will share an access to
            only records in the `prefix` key range, but with `prefix`-es
            stripped from them.

            For example, if the Container contents are:
            `{
                b'e.y': b'012',
                b'meta.x': b'123',
                b'meta.z': b'x',
                b'zzz': b'oOo'
            }`, then `container.view(prefix=b'meta.')` will behave (almost)
            exactly as an Container:
            `{
                b'x': b'123',
                b'z': b'x',
            }`
        container (:obj:`Container`): the parent container to build the view on
    """

    def __init__(
        self,
        *,
        prefix: bytes = b'',
        container: Container,
        read_only: bool = None
    ) -> None:
        self.prefix = prefix
        self.parent = container

        self.read_only = read_only

    def close(self):
        """Close all the resources."""
        pass

    def preload(self):
        """Preload the Container.

        The interface of Container is designed in such a way that (almost) all
        the operations are supported to be done lazily.
        Sometimes there is need to preload the storage without performing an
        operation that will cause an actual read / write access.
        """
        self.parent.preload()

    def finalize(self, index: Container):
        """Finalize the Container.

        Perform operations of compactions, indexing, optimization, etc.
        """
        prefix = self.absolute_path()
        # Shadowing
        index.delete_range(prefix, prefix + b'\xff')  # TODO check validity
        self.parent.finalize(index=index)

    def absolute_path(
        self,
        path: bytes = None
    ) -> bytes:
        """Returns the absolute path for the given relative `path`.

        Path separators / sentinels should be handled in higher level so that
        `join(a, b) == a + b` property holds. This can be easily achieved by
        having all the paths end with the sentinel:
        `join('a/b/c/', 'e/f/') == 'a/b/c/' + 'e/f/' = 'a/b/c/e/f/'`
        """
        if path is None:
            return self.prefix
        return self.prefix + path

    def get(
        self,
        key: ContainerKey,
        default=None
    ) -> ContainerValue:
        """Returns the value by the given `key` if it exists else `default`.

        The `default` is :obj:`None` by default.
        """
        path = self.absolute_path(key)
        return self.parent.get(path, default)

    def __getitem__(
        self,
        key: ContainerKey
    ) -> ContainerValue:
        """Returns the value by the given `key`."""
        path = self.absolute_path(key)
        return self.parent[path]

    def set(
        self,
        key: ContainerKey,
        value: ContainerValue,
        store_batch=None
    ) -> None:
        """Set a value for given key, optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations. Depending
        on the :obj:`Conainer` implementation, this may feature transactions,
        atomic writes, etc.
        """
        path = self.absolute_path(key)
        self.parent.set(path, value, store_batch=store_batch)

    def __setitem__(
        self,
        key: ContainerKey,
        value: ContainerValue
    ) -> None:
        """Set a value for given key."""
        path = self.absolute_path(key)
        self.parent[path] = value

    def delete(
        self,
        key: ContainerKey,
        store_batch=None
    ):
        """Delete a key-value record by the given key,
        optionally store in a batch.

        If `store_batch` is provided, instead of the `(key, value)` being added
        to the collection immediately, the operation is stored in a batch in
        order to be executed in a whole with other write operations. Depending
        on the :obj:`Conainer` implementation, this may feature transactions,
        atomic writes, etc.
        """
        path = self.absolute_path(key)
        return self.parent.delete(path, store_batch=store_batch)

    def __delitem__(
        self,
        key: ContainerKey
    ) -> None:
        """Delete a key-value record by the given key."""
        path = self.absolute_path(key)
        del self.parent[path]

    def delete_range(
        self,
        begin: ContainerKey,
        end: ContainerKey,
        store_batch=None
    ):
        """Delete all the records in the given `[begin, end)` key range."""
        begin_path = self.absolute_path(begin)
        end_path = self.absolute_path(end)
        self.parent.delete_range(begin_path, end_path,
                                 store_batch=store_batch)

    def next_item(
        self,
        key: ContainerKey = b''
    ) -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right after the provided `key`.
        """
        path = self.absolute_path(key)
        keys, value = self.parent.next_item(path)
        if path:
            _prefix, _path, keys = keys.partition(path)
            if _prefix or _path != path:
                raise KeyError
        return keys, value

    def prev_item(
        self,
        key: ContainerKey = b''
    ) -> Tuple[ContainerKey, ContainerValue]:
        """Returns `(key, value)` for the key that comes (lexicographically)
        right before the provided `key`.
        """
        path = self.absolute_path(key)
        keys, value = self.parent.prev_item(path)
        if path:
            _prefix, _path, keys = keys.partition(path)
            if _prefix or _path != path:
                raise KeyError
        return keys, value

    def walk(
        self,
        key: ContainerKey = b''
    ):
        """A bi-directional generator to walk over the collection of records on
        any arbitrary order. The `prefix` sent to the generator (lets call it
        a `walker`) seeks for lower-bound key in the collection.

        In other words, if the Container contents are:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }` and `walker = container.walk()` then:
        `walker.send(b'meta') == b'meta.x'`, `walker.send(b'e.y') == b'e.y'`
        """
        path = self.absolute_path(key)
        walker = self.parent.walk(path)
        p = None
        while True:
            if p is None:
                next_key = next(walker)
            else:
                next_key = walker.send(p)
            if next_key is None:
                return
            if path:
                _prefix, _path, next_key = next_key.partition(path)
                # assert not _prefix and _path == path
                if _prefix or _path != path:
                    return

            key = yield next_key
            p = self.absolute_path(key)

    def items(
        self,
        key: ContainerKey = b''
    ) -> Iterator[Tuple[ContainerKey, ContainerValue]]:
        """Iterate over all the key-value records in the prefix key range.

        The iteration is always performed in lexiographic order w.r.t keys.
        If `prefix` is provided, iterate only over those records that have key
        starting with the `prefix`.

        For example, if `prefix == b'meta.'`, and the Container consists of:
        `{
            b'e.y': b'012',
            b'meta.x': b'123',
            b'meta.z': b'x',
            b'zzz': b'oOo'
        }`, the method will yield `(b'meta.x', b'123')` and `(b'meta.z', b'x')`

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range
        """
        return PrefixViewItemsIterator(self, key)

    def view(
        self,
        prefix: bytes = b''
    ) -> Container:
        """Return a view (even mutable ones) that enable access to the container
        but with modifications.

        Args:
            prefix (:obj:`bytes`): the prefix that defines the key range of the
                view-container. The resulting container will share an access to
                only records in the `prefix` key range, but with `prefix`-es
                stripped from them.

                For example, if the Container contents are:
                `{
                    b'e.y': b'012',
                    b'meta.x': b'123',
                    b'meta.z': b'x',
                    b'zzz': b'oOo'
                }`, then `container.view(prefix=b'meta.')` will behave (almost)
                exactly as an Container:
                `{
                    b'x': b'123',
                    b'z': b'x',
                }`
        """
        # TODO *args instead?
        return self.parent.view(self.prefix + prefix)

    def tree(
        self
    ) -> ContainerTreeView:
        """Return a :obj:`ContainerTreeView` which enables hierarchical view and access
        to the container records.

        This is achieved by prefixing groups and using `PATH_SENTINEL` as a
        separator for keys.

        For example, if the Container contents are:
            `{
                b'e.y': b'012',
                b'meta.x': b'123',
                b'meta.z': b'x',
                b'zzz': b'oOo'
            }`, and the path sentinel is `b'.'` then `tree = container.tree()`
            will behave as a (possibly deep) dict-like object:
            `tree[b'meta'][b'x'] == b'123'`
        """
        return ContainerTreeView(self)

    def batch(self):
        """Creates a new batch object to store operations in before executing
        using :obj:`Container.commit`.

        The operations :obj:`Container.set`, :obj:`Container.delete`,
        :obj:`Container.delete_range` are supported.

        See more at :obj:`Container.commit`
        """
        return self.parent.batch()

    def commit(
        self,
        batch
    ):
        """Execute the accumulated write operations in the given `batch`.

        Depending on the :obj:`Container` implementation, this may feature
        transactions, atomic writes, etc.
        """
        return self.parent.commit(batch)


class PrefixViewItemsIterator(ContainerItemsIterator):

    def __init__(self, prefix_view, key):
        self.prefix_view = prefix_view
        self.path = prefix_view.absolute_path(key)
        self.it = prefix_view.parent.items(self.path)
        # We store length of the path to strip it from the key faster
        self.prefix_len = len(self.path)

    def next(self):
        item = self.it.next()

        if item is None:
            return None

        keys = item[0]
        value = item[1]

        return keys[self.prefix_len:], value



"""empty message

Revision ID: 11672b13f92c
Revises: 
Create Date: 2021-05-18 15:06:30.858112

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '11672b13f92c'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('commits',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('hash', sa.Text(), nullable=True),
    sa.Column('experiment_name', sa.Text(), nullable=True),
    sa.Column('session_started_at', sa.Integer(), nullable=True),
    sa.Column('session_closed_at', sa.Integer(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('uuid'),
    sa.UniqueConstraint('experiment_name', 'hash')
    )
    op.create_table('tags',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('name', sa.Text(), nullable=True),
    sa.Column('color', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.Column('is_hidden', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('tf_summary_logs',
    sa.Column('uuid', sa.Text(), nullable=False),
    sa.Column('log_path', sa.Text(), nullable=True),
    sa.Column('params', sa.Text(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('is_archived', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('uuid')
    )
    op.create_table('commit_tag',
    sa.Column('commit_id', sa.Text(), nullable=True),
    sa.Column('tag_id', sa.Text(), nullable=True),
    sa.ForeignKeyConstraint(['commit_id'], ['commits.uuid'], ),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.uuid'], )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('commit_tag')
    op.drop_table('tf_summary_logs')
    op.drop_table('tags')
    op.drop_table('commits')
    # ### end Alembic commands ###



# distutils: language = c++
# cython: language_level = 3

"""Hashing implementations of AimObject-compatible data.
This is different from CPython's implementation because of many reasons:
  *  CPython uses non-deterministic hashing to avoid collision attacks.
     Although we also cover such scenarios, in contrast to CPython, we do not
     use PYTHONHASHSEED environment variable.
  *  We implement hash for mutable objects as well (`list`, `dict`, etc.)
  *  Our implementation are less prone to manually designed collisions.
"""

ctypedef long long int64

import hashlib

from aim.storage.encoding import encode_int64, decode_int64  # noqa
from aim.storage.hashing cimport c_hash

cdef int64 hash_none(object obj = *)
cdef int64 hash_uniform(int64 bad_hash)
cdef int64 hash_number(double number)
cdef int64 hash_bool(bint obj)
cdef int64 hash_bytes(bytes obj)
cdef int64 hash_string(str obj)
cdef int64 hash_array(obj)
cdef int64 key_cmp(tuple obj)
cdef int64 hash_object(obj)
cpdef int64 hash_auto(obj)



# distutils: language = c++
# cython: language_level = 3

from aim.storage.arrayview cimport ArrayView
from aim.storage.treeview cimport TreeView

cdef class TreeArrayView(ArrayView):
    cdef:
        public TreeView tree
        public object dtype
